# Copyright (c) The treq Authors.
# See LICENSE for details.
from twisted.trial.unittest import SynchronousTestCase
from twisted.web.http_headers import Headers
from twisted.web.iweb import IAgent

from treq._agentspy import agent_spy
from treq.auth import _RequestHeaderSetterAgent, add_auth, UnknownAuthConfig


class RequestHeaderSetterAgentTests(SynchronousTestCase):
    def setUp(self):
        self.agent, self.requests = agent_spy()

    def test_sets_headers(self):
        agent = _RequestHeaderSetterAgent(
            self.agent,
            Headers({b'X-Test-Header': [b'Test-Header-Value']}),
        )

        agent.request(b'method', b'uri')

        self.assertEqual(
            self.requests[0].headers,
            Headers({b'X-Test-Header': [b'Test-Header-Value']}),
        )

    def test_overrides_per_request_headers(self):
        agent = _RequestHeaderSetterAgent(
            self.agent,
            Headers({b'X-Test-Header': [b'Test-Header-Value']})
        )

        agent.request(
            b'method', b'uri',
            Headers({b'X-Test-Header': [b'Unwanted-Value']})
        )

        self.assertEqual(
            self.requests[0].headers,
            Headers({b'X-Test-Header': [b'Test-Header-Value']}),
        )

    def test_no_mutation(self):
        """
        The agent never mutates the headers passed to its request method.

        This reproduces https://github.com/twisted/treq/issues/314
        """
        requestHeaders = Headers({})
        agent = _RequestHeaderSetterAgent(
            self.agent,
            Headers({b'Added': [b'1']}),
        )

        agent.request(b'method', b'uri', headers=requestHeaders)

        self.assertEqual(requestHeaders, Headers({}))


class AddAuthTests(SynchronousTestCase):
    def test_add_basic_auth(self):
        """
        add_auth() wraps the given agent with one that adds an ``Authorization:
        Basic ...`` HTTP header that contains the given credentials.
        """
        agent, requests = agent_spy()
        authAgent = add_auth(agent, ('username', 'password'))

        authAgent.request(b'method', b'uri')

        self.assertTrue(IAgent.providedBy(authAgent))
        self.assertEqual(
            requests[0].headers,
            Headers({b'authorization': [b'Basic dXNlcm5hbWU6cGFzc3dvcmQ=']})
        )

    def test_add_basic_auth_huge(self):
        """
        The Authorization header doesn't include linebreaks, even if the
        credentials are so long that Python's base64 implementation inserts
        them.
        """
        agent, requests = agent_spy()
        pwd = ('verylongpasswordthatextendsbeyondthepointwheremultiplel'
               'inesaregenerated')
        expectedAuth = (
            b'Basic dXNlcm5hbWU6dmVyeWxvbmdwYXNzd29yZHRoYXRleHRlbmRzY'
            b'mV5b25kdGhlcG9pbnR3aGVyZW11bHRpcGxlbGluZXNhcmVnZW5lcmF0ZWQ='
        )
        authAgent = add_auth(agent, ('username', pwd))

        authAgent.request(b'method', b'uri')

        self.assertEqual(
            requests[0].headers,
            Headers({b'authorization': [expectedAuth]}),
        )

    def test_add_basic_auth_utf8(self):
        """
        Basic auth username and passwords given as `str` are encoded as UTF-8.

        https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication#Character_encoding_of_HTTP_authentication
        """
        agent, requests = agent_spy()
        auth = (u'\u16d7', u'\u16b9')
        authAgent = add_auth(agent, auth)

        authAgent.request(b'method', b'uri')

        self.assertEqual(
            requests[0].headers,
            Headers({b'Authorization': [b'Basic 4ZuXOuGauQ==']}),
        )

    def test_add_basic_auth_bytes(self):
        """
        Basic auth can be passed as `bytes`, allowing the user full control
        over the encoding.
        """
        agent, requests = agent_spy()
        auth = (b'\x01\x0f\xff', b'\xff\xf0\x01')
        authAgent = add_auth(agent, auth)

        authAgent.request(b'method', b'uri')

        self.assertEqual(
            requests[0].headers,
            Headers({b'Authorization': [b'Basic AQ//Ov/wAQ==']}),
        )

    def test_add_unknown_auth(self):
        """
        add_auth() raises UnknownAuthConfig when given anything other than
        a tuple.
        """
        agent, _ = agent_spy()
        invalidAuth = 1234

        self.assertRaises(UnknownAuthConfig, add_auth, agent, invalidAuth)



from treq.client import HTTPClient
from _utils import print_response
from twisted.internet.task import react
from twisted.web.client import Agent

def make_custom_agent(reactor):
    return Agent(reactor, connectTimeout=42)

def main(reactor, *args):
    agent = make_custom_agent(reactor)
    http_client = HTTPClient(agent)
    d = http_client.get(
        'https://secure.example.net/area51',
        auth=('admin', "you'll never guess!"))
    d.addCallback(print_response)
    return d

react(main, [])




"""
Tests for :py:mod:`treq.test.local_httpbin.child`
"""
import attr

from cryptography.hazmat.primitives.asymmetric import padding

import functools

import io

from twisted.trial.unittest import SynchronousTestCase

try:
    from twisted.internet.testing import MemoryReactor
except ImportError:
    from twisted.test.proto_helpers import MemoryReactor

from twisted.internet import defer

from treq.test.util import skip_on_windows_because_of_199

from twisted.web.server import Site
from twisted.web.resource import Resource

from service_identity.cryptography import verify_certificate_hostname

from .. import child, shared


skip = skip_on_windows_because_of_199()


class CertificatesForAuthorityAndServerTests(SynchronousTestCase):
    """
    Tests for :py:func:`child._certificates_for_authority_and_server`
    """

    def setUp(self):
        self.hostname = u".example.org"
        (
            self.ca_cert,
            self.server_private_key,
            self.server_x509_cert,
        ) = child._certificates_for_authority_and_server(
            self.hostname,
        )

    def test_pkey_x509_paired(self):
        """
        The returned private key corresponds to the X.509
        certificate's public key.
        """
        server_private_key = self.server_private_key.to_cryptography_key()
        server_x509_cert = self.server_x509_cert.to_cryptography()

        plaintext = b'plaintext'
        ciphertext = server_x509_cert.public_key().encrypt(
            plaintext,
            padding.PKCS1v15(),
        )

        self.assertEqual(
            server_private_key.decrypt(
                ciphertext,
                padding.PKCS1v15(),
            ),
            plaintext,
        )

    def test_ca_signed_x509(self):
        """
        The returned X.509 certificate was signed by the returned
        certificate authority's certificate.
        """
        ca_cert = self.ca_cert.original.to_cryptography()
        server_x509_cert = self.server_x509_cert.to_cryptography()

        # Raises an InvalidSignature exception on failure.
        ca_cert.public_key().verify(
            server_x509_cert.signature,
            server_x509_cert.tbs_certificate_bytes,
            padding.PKCS1v15(),
            server_x509_cert.signature_hash_algorithm
        )

    def test_x509_matches_hostname(self):
        """
        The returned X.509 certificate is valid for the hostname.
        """
        verify_certificate_hostname(
            self.server_x509_cert.to_cryptography(),
            self.hostname,
        )


@attr.s
class FakeThreadPoolState(object):
    """
    State for :py:class:`FakeThreadPool`.
    """
    init_call_count = attr.ib(default=0)
    start_call_count = attr.ib(default=0)


@attr.s
class FakeThreadPool(object):
    """
    A fake :py:class:`twisted.python.threadpool.ThreadPool`
    """
    _state = attr.ib()

    def init(self):
        self._state.init_call_count += 1
        return self

    def start(self):
        """
        See :py:meth:`twisted.python.threadpool.ThreadPool.start`
        """
        self._state.start_call_count += 1

    def stop(self):
        """
        See :py:meth:`twisted.python.threadpool.ThreadPool.stop`
        """


class MakeHTTPBinSiteTests(SynchronousTestCase):
    """
    Tests for :py:func:`_make_httpbin_site`.
    """

    def setUp(self):
        self.fake_threadpool_state = FakeThreadPoolState()
        self.fake_threadpool = FakeThreadPool(self.fake_threadpool_state)
        self.reactor = MemoryReactor()

    def test_threadpool_management(self):
        """
        A thread pool is created that will be shut down when the
        reactor shuts down.
        """
        child._make_httpbin_site(
            self.reactor,
            threadpool_factory=self.fake_threadpool.init,
        )

        self.assertEqual(self.fake_threadpool_state.init_call_count, 1)
        self.assertEqual(self.fake_threadpool_state.start_call_count, 1)

        self.assertEqual(len(self.reactor.triggers['before']['shutdown']), 1)
        [(stop, _, _)] = self.reactor.triggers['before']['shutdown']

        self.assertEqual(stop, self.fake_threadpool.stop)


class ServeTLSTests(SynchronousTestCase):
    """
    Tests for :py:func:`_serve_tls`
    """

    def setUp(self):
        self.reactor = MemoryReactor()
        self.site = Site(Resource())

    def test_tls_listener_matches_description(self):
        """
        An SSL listener is established on the requested host and port,
        and the host, port, and CA certificate are returned in its
        description.
        """
        expected_host = 'host'
        expected_port = 123

        description_deferred = child._serve_tls(
            self.reactor,
            host=expected_host,
            port=expected_port,
            site=self.site,
        )

        self.assertEqual(len(self.reactor.sslServers), 1)

        [
            (actual_port, actual_site, _, _, actual_host)
        ] = self.reactor.sslServers

        self.assertEqual(actual_host, expected_host)
        self.assertEqual(actual_port, expected_port)
        self.assertIs(actual_site, self.site)

        description = self.successResultOf(description_deferred)

        self.assertEqual(description.host, expected_host)
        self.assertEqual(description.port, expected_port)
        self.assertTrue(description.cacert)


class ServeTCPTests(SynchronousTestCase):
    """
    Tests for :py:func:`_serve_tcp`
    """

    def setUp(self):
        self.reactor = MemoryReactor()
        self.site = Site(Resource)

    def test_tcp_listener_matches_description(self):
        """
        A TCP listeneris established on the request host and port, and
        the host and port are returned in its description.
        """
        expected_host = 'host'
        expected_port = 123

        description_deferred = child._serve_tcp(
            self.reactor,
            host=expected_host,
            port=expected_port,
            site=self.site,
        )

        self.assertEqual(len(self.reactor.tcpServers), 1)

        [
            (actual_port, actual_site, _, actual_host)
        ] = self.reactor.tcpServers

        self.assertEqual(actual_host, expected_host)
        self.assertEqual(actual_port, expected_port)
        self.assertIs(actual_site, self.site)

        description = self.successResultOf(description_deferred)

        self.assertEqual(description.host, expected_host)
        self.assertEqual(description.port, expected_port)
        self.assertFalse(description.cacert)


@attr.s
class FlushableBytesIOState(object):
    """
    State for :py:class:`FlushableBytesIO`
    """
    bio = attr.ib(default=attr.Factory(io.BytesIO))
    flush_count = attr.ib(default=0)


@attr.s
class FlushableBytesIO(object):
    """
    A :py:class:`io.BytesIO` wrapper that records flushes.
    """
    _state = attr.ib()

    def write(self, data):
        self._state.bio.write(data)

    def flush(self):
        self._state.flush_count += 1


@attr.s
class BufferedStandardOut(object):
    """
    A standard out that whose ``buffer`` is a
    :py:class:`FlushableBytesIO` instance.
    """
    buffer = attr.ib()


class OutputProcessDescriptionTests(SynchronousTestCase):
    """
    Tests for :py:func:`_output_process_description`
    """

    def setUp(self):
        self.stdout_state = FlushableBytesIOState()
        self.stdout = BufferedStandardOut(FlushableBytesIO(self.stdout_state))

    def test_description_written(self):
        """
        An :py:class:`shared._HTTPBinDescription` is written to
        standard out and the line flushed.
        """
        description = shared._HTTPBinDescription(host="host",
                                                 port=123,
                                                 cacert="cacert")

        child._output_process_description(description, self.stdout)

        written = self.stdout_state.bio.getvalue()

        self.assertEqual(
            written,
            b'{"cacert": "cacert", "host": "host", "port": 123}' + b'\n',
        )

        self.assertEqual(self.stdout_state.flush_count, 1)


class ForeverHTTPBinTests(SynchronousTestCase):
    """
    Tests for :py:func:`_forever_httpbin`
    """

    def setUp(self):
        self.make_httpbin_site_returns = Site(Resource())

        self.serve_tcp_calls = []
        self.serve_tcp_returns = defer.Deferred()

        self.serve_tls_calls = []
        self.serve_tls_returns = defer.Deferred()

        self.output_process_description_calls = []
        self.output_process_description_returns = None

        self.reactor = MemoryReactor()

        self.forever_httpbin = functools.partial(
            child._forever_httpbin,
            _make_httpbin_site=self.make_httpbin_site,
            _serve_tcp=self.serve_tcp,
            _serve_tls=self.serve_tls,
            _output_process_description=self.output_process_description,
        )

    def make_httpbin_site(self, reactor, *args, **kwargs):
        """
        A fake :py:func:`child._make_httpbin_site`.
        """
        return self.make_httpbin_site_returns

    def serve_tcp(self, reactor, host, port, site):
        """
        A fake :py:func:`child._serve_tcp`.
        """
        self.serve_tcp_calls.append((reactor, host, port, site))
        return self.serve_tcp_returns

    def serve_tls(self, reactor, host, port, site):
        """
        A fake :py:func:`child._serve_tls`.
        """
        self.serve_tls_calls.append((reactor, host, port, site))
        return self.serve_tls_returns

    def output_process_description(self, description, *args, **kwargs):
        """
        A fake :py:func:`child._output_process_description`
        """
        self.output_process_description_calls.append(description)
        return self.output_process_description_returns

    def assertDescriptionAndDeferred(self,
                                     description_deferred,
                                     forever_deferred):
        """
        Assert that firing ``description_deferred`` outputs the
        description but that ``forever_deferred`` never fires.
        """
        description_deferred.callback("description")

        self.assertEqual(self.output_process_description_calls,
                         ["description"])

        self.assertNoResult(forever_deferred)

    def test_default_arguments(self):
        """
        The default command line arguments host ``httpbin`` on
        ``localhost`` and a randomly-assigned port, returning a
        :py:class:`Deferred` that never fires.
        """
        deferred = self.forever_httpbin(self.reactor, [])

        self.assertEqual(
            self.serve_tcp_calls,
            [
                (self.reactor, 'localhost', 0, self.make_httpbin_site_returns)
            ]
        )

        self.assertDescriptionAndDeferred(
            description_deferred=self.serve_tcp_returns,
            forever_deferred=deferred,
        )

    def test_https(self):
        """
        The ``--https`` command line argument serves ``httpbin`` over
        HTTPS, returning a :py:class:`Deferred` that never fires.
        """
        deferred = self.forever_httpbin(self.reactor, ['--https'])

        self.assertEqual(
            self.serve_tls_calls,
            [
                (self.reactor, 'localhost', 0, self.make_httpbin_site_returns)
            ]
        )

        self.assertDescriptionAndDeferred(
            description_deferred=self.serve_tls_returns,
            forever_deferred=deferred,
        )

    def test_host(self):
        """
        The ``--host`` command line argument serves ``httpbin`` on
        provided host, returning a :py:class:`Deferred` that never
        fires.
        """
        deferred = self.forever_httpbin(self.reactor,
                                        ['--host', 'example.org'])

        self.assertEqual(
            self.serve_tcp_calls,
            [
                (
                    self.reactor,
                    'example.org',
                    0,
                    self.make_httpbin_site_returns,
                )
            ]
        )

        self.assertDescriptionAndDeferred(
            description_deferred=self.serve_tcp_returns,
            forever_deferred=deferred,
        )

    def test_port(self):
        """
        The ``--port`` command line argument serves ``httpbin`` on
        the provided port, returning a :py:class:`Deferred` that never
        fires.
        """
        deferred = self.forever_httpbin(self.reactor, ['--port', '91'])

        self.assertEqual(
            self.serve_tcp_calls,
            [
                (self.reactor, 'localhost', 91, self.make_httpbin_site_returns)
            ]
        )

        self.assertDescriptionAndDeferred(
            description_deferred=self.serve_tcp_returns,
            forever_deferred=deferred,
        )



from decimal import Decimal

from twisted.trial.unittest import SynchronousTestCase

from twisted.python.failure import Failure
from twisted.web.client import ResponseDone
from twisted.web.iweb import UNKNOWN_LENGTH
from twisted.web.http_headers import Headers

from treq.response import _Response


class FakeResponse:
    def __init__(self, code, headers, body=()):
        self.code = code
        self.headers = headers
        self.previousResponse = None
        self._body = body
        self.length = sum(len(c) for c in body)

    def setPreviousResponse(self, response):
        self.previousResponse = response

    def deliverBody(self, protocol):
        for chunk in self._body:
            protocol.dataReceived(chunk)
        protocol.connectionLost(Failure(ResponseDone()))


class ResponseTests(SynchronousTestCase):
    def test_repr_content_type(self):
        """
        When the response has a Content-Type header its value is included in
        the response.
        """
        headers = Headers({'Content-Type': ['text/html']})
        original = FakeResponse(200, headers, body=[b'<!DOCTYPE html>'])
        self.assertEqual(
            "<treq.response._Response 200 'text/html' 15 bytes>",
            repr(_Response(original, None)),
        )

    def test_repr_content_type_missing(self):
        """
        A request with no Content-Type just displays an empty field.
        """
        original = FakeResponse(204, Headers(), body=[b''])
        self.assertEqual(
            "<treq.response._Response 204 '' 0 bytes>",
            repr(_Response(original, None)),
        )

    def test_repr_content_type_hostile(self):
        """
        Garbage in the Content-Type still produces a reasonable representation.
        """
        headers = Headers({'Content-Type': [u'\u2e18', ' x/y']})
        original = FakeResponse(418, headers, body=[b''])
        self.assertEqual(
            r"<treq.response._Response 418 '\xe2\xb8\x98,  x/y' 0 bytes>",
            repr(_Response(original, None)),
        )

    def test_repr_unknown_length(self):
        """
        A HTTP 1.0 or chunked response displays an unknown length.
        """
        headers = Headers({'Content-Type': ['text/event-stream']})
        original = FakeResponse(200, headers)
        original.length = UNKNOWN_LENGTH
        self.assertEqual(
            "<treq.response._Response 200 'text/event-stream' unknown size>",
            repr(_Response(original, None)),
        )

    def test_collect(self):
        original = FakeResponse(200, Headers(), body=[b'foo', b'bar', b'baz'])
        calls = []
        _Response(original, None).collect(calls.append)
        self.assertEqual([b'foo', b'bar', b'baz'], calls)

    def test_content(self):
        original = FakeResponse(200, Headers(), body=[b'foo', b'bar', b'baz'])
        self.assertEqual(
            b'foobarbaz',
            self.successResultOf(_Response(original, None).content()),
        )

    def test_json(self):
        original = FakeResponse(200, Headers(), body=[b'{"foo": ', b'"bar"}'])
        self.assertEqual(
            {'foo': 'bar'},
            self.successResultOf(_Response(original, None).json()),
        )

    def test_json_customized(self):
        original = FakeResponse(200, Headers(), body=[b'{"foo": ',
                                                      b'1.0000000000000001}'])
        self.assertEqual(
            self.successResultOf(_Response(original, None).json(
                parse_float=Decimal)
            )["foo"],
            Decimal("1.0000000000000001")
        )

    def test_text(self):
        headers = Headers({b'content-type': [b'text/plain;charset=utf-8']})
        original = FakeResponse(200, headers, body=[b'\xe2\x98', b'\x83'])
        self.assertEqual(
            u'\u2603',
            self.successResultOf(_Response(original, None).text()),
        )

    def test_history(self):
        redirect1 = FakeResponse(
            301,
            Headers({'location': ['http://example.com/']})
        )

        redirect2 = FakeResponse(
            302,
            Headers({'location': ['https://example.com/']})
        )
        redirect2.setPreviousResponse(redirect1)

        final = FakeResponse(200, Headers({}))
        final.setPreviousResponse(redirect2)

        wrapper = _Response(final, None)

        history = wrapper.history()

        self.assertEqual(wrapper.code, 200)
        self.assertEqual(history[0].code, 301)
        self.assertEqual(history[1].code, 302)

    def test_no_history(self):
        wrapper = _Response(FakeResponse(200, Headers({})), None)
        self.assertEqual(wrapper.history(), [])



from twisted.internet.task import react
from _utils import print_response

import treq


def main(reactor):
    d = treq.post("https://httpbin.org/post",
                  data={"form": "data"})
    d.addCallback(print_response)
    return d

react(main, [])



from twisted.internet.task import react

import treq


def download_file(reactor, url, destination_filename):
    destination = open(destination_filename, 'wb')
    d = treq.get(url, unbuffered=True)
    d.addCallback(treq.collect, destination.write)
    d.addBoth(lambda _: destination.close())
    return d

react(download_file, ['https://httpbin.org/get', 'download.txt'])



import os
import platform

from unittest import mock

from twisted.internet import reactor
from twisted.internet.task import Clock

DEBUG = os.getenv("TREQ_DEBUG", False) == "true"

is_pypy = platform.python_implementation() == 'PyPy'


def with_clock(fn):
    def wrapper(*args, **kwargs):
        clock = Clock()
        with mock.patch.object(reactor, 'callLater', clock.callLater):
            return fn(*(args + (clock,)), **kwargs)
    return wrapper


def skip_on_windows_because_of_199():
    """
    Return a skip describing issue #199 under Windows.

    :return: A :py:class:`str` skip reason.
    """
    if platform.system() == 'Windows':
        return ("HTTPBin process cannot run under Windows."
                " See https://github.com/twisted/treq/issues/199")
    return None



import json

from zope.interface import implementer
from twisted.web.resource import IResource


@implementer(IResource)
class JsonResource(object):
    isLeaf = True  # NB: means getChildWithDefault will not be called

    def __init__(self, data):
        self.data = data

    def render(self, request):
        request.setHeader(b'Content-Type', b'application/json')
        return json.dumps(self.data).encode('utf-8')



import io
import mimetypes
import uuid
import warnings
from collections.abc import Mapping
from http.cookiejar import CookieJar
from urllib.parse import quote_plus, urlencode as _urlencode

from twisted.internet.interfaces import IProtocol
from twisted.internet.defer import Deferred
from twisted.python.components import proxyForInterface
from twisted.python.filepath import FilePath
from hyperlink import DecodedURL, EncodedURL

from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer, IResponse

from twisted.web.client import (
    FileBodyProducer,
    RedirectAgent,
    BrowserLikeRedirectAgent,
    ContentDecoderAgent,
    GzipDecoder,
    CookieAgent
)

from twisted.python.components import registerAdapter
from json import dumps as json_dumps

from treq.auth import add_auth
from treq import multipart
from treq.response import _Response
from requests.cookies import cookiejar_from_dict, merge_cookies


_NOTHING = object()


def urlencode(query, doseq):
    s = _urlencode(query, doseq)
    if not isinstance(s, bytes):
        s = s.encode("ascii")
    return s


class _BodyBufferingProtocol(proxyForInterface(IProtocol)):
    def __init__(self, original, buffer, finished):
        self.original = original
        self.buffer = buffer
        self.finished = finished

    def dataReceived(self, data):
        self.buffer.append(data)
        self.original.dataReceived(data)

    def connectionLost(self, reason):
        self.original.connectionLost(reason)
        self.finished.errback(reason)


class _BufferedResponse(proxyForInterface(IResponse)):
    def __init__(self, original):
        self.original = original
        self._buffer = []
        self._waiters = []
        self._waiting = None
        self._finished = False
        self._reason = None

    def _deliverWaiting(self, reason):
        self._reason = reason
        self._finished = True
        for waiter in self._waiters:
            for segment in self._buffer:
                waiter.dataReceived(segment)
            waiter.connectionLost(reason)

    def deliverBody(self, protocol):
        if self._waiting is None and not self._finished:
            self._waiting = Deferred()
            self._waiting.addBoth(self._deliverWaiting)
            self.original.deliverBody(
                _BodyBufferingProtocol(
                    protocol,
                    self._buffer,
                    self._waiting
                )
            )
        elif self._finished:
            for segment in self._buffer:
                protocol.dataReceived(segment)
            protocol.connectionLost(self._reason)
        else:
            self._waiters.append(protocol)


class HTTPClient:
    def __init__(self, agent, cookiejar=None,
                 data_to_body_producer=IBodyProducer):
        self._agent = agent
        self._cookiejar = cookiejar or cookiejar_from_dict({})
        self._data_to_body_producer = data_to_body_producer

    def get(self, url, **kwargs):
        """
        See :func:`treq.get()`.
        """
        kwargs.setdefault('_stacklevel', 3)
        return self.request('GET', url, **kwargs)

    def put(self, url, data=None, **kwargs):
        """
        See :func:`treq.put()`.
        """
        kwargs.setdefault('_stacklevel', 3)
        return self.request('PUT', url, data=data, **kwargs)

    def patch(self, url, data=None, **kwargs):
        """
        See :func:`treq.patch()`.
        """
        kwargs.setdefault('_stacklevel', 3)
        return self.request('PATCH', url, data=data, **kwargs)

    def post(self, url, data=None, **kwargs):
        """
        See :func:`treq.post()`.
        """
        kwargs.setdefault('_stacklevel', 3)
        return self.request('POST', url, data=data, **kwargs)

    def head(self, url, **kwargs):
        """
        See :func:`treq.head()`.
        """
        kwargs.setdefault('_stacklevel', 3)
        return self.request('HEAD', url, **kwargs)

    def delete(self, url, **kwargs):
        """
        See :func:`treq.delete()`.
        """
        kwargs.setdefault('_stacklevel', 3)
        return self.request('DELETE', url, **kwargs)

    def request(
        self,
        method,
        url,
        *,
        params=None,
        headers=None,
        data=None,
        files=None,
        json=_NOTHING,
        auth=None,
        cookies=None,
        allow_redirects=True,
        browser_like_redirects=False,
        unbuffered=False,
        reactor=None,
        timeout=None,
        _stacklevel=2,
    ):
        """
        See :func:`treq.request()`.
        """
        method = method.encode('ascii').upper()

        if isinstance(url, DecodedURL):
            parsed_url = url.encoded_url
        elif isinstance(url, EncodedURL):
            parsed_url = url
        elif isinstance(url, str):
            # We use hyperlink in lazy mode so that users can pass arbitrary
            # bytes in the path and querystring.
            parsed_url = EncodedURL.from_text(url)
        else:
            parsed_url = EncodedURL.from_text(url.decode('ascii'))

        # Join parameters provided in the URL
        # and the ones passed as argument.
        if params:
            parsed_url = parsed_url.replace(
                query=parsed_url.query + tuple(_coerced_query_params(params))
            )

        url = parsed_url.to_uri().to_text().encode('ascii')

        headers = self._request_headers(headers, _stacklevel + 1)

        bodyProducer, contentType = self._request_body(data, files, json,
                                                       stacklevel=_stacklevel + 1)
        if contentType is not None:
            headers.setRawHeaders(b'Content-Type', [contentType])

        if not isinstance(cookies, CookieJar):
            cookies = cookiejar_from_dict(cookies)

        cookies = merge_cookies(self._cookiejar, cookies)
        wrapped_agent = CookieAgent(self._agent, cookies)

        if allow_redirects:
            if browser_like_redirects:
                wrapped_agent = BrowserLikeRedirectAgent(wrapped_agent)
            else:
                wrapped_agent = RedirectAgent(wrapped_agent)

        wrapped_agent = ContentDecoderAgent(wrapped_agent,
                                            [(b'gzip', GzipDecoder)])

        if auth:
            wrapped_agent = add_auth(wrapped_agent, auth)

        d = wrapped_agent.request(
            method, url, headers=headers,
            bodyProducer=bodyProducer)

        if reactor is None:
            from twisted.internet import reactor
        if timeout:
            delayedCall = reactor.callLater(timeout, d.cancel)

            def gotResult(result):
                if delayedCall.active():
                    delayedCall.cancel()
                return result

            d.addBoth(gotResult)

        if not unbuffered:
            d.addCallback(_BufferedResponse)

        return d.addCallback(_Response, cookies)

    def _request_headers(self, headers, stacklevel):
        """
        Convert the *headers* argument to a :class:`Headers` instance

        :returns:
            :class:`twisted.web.http_headers.Headers`
        """
        if isinstance(headers, dict):
            h = Headers({})
            for k, v in headers.items():
                if isinstance(v, (bytes, str)):
                    h.addRawHeader(k, v)
                elif isinstance(v, list):
                    h.setRawHeaders(k, v)
                else:
                    warnings.warn(
                        (
                            "The value of headers key {!r} has non-string type {}"
                            " and will be dropped."
                            " This will raise TypeError in the next treq release."
                        ).format(k, type(v)),
                        DeprecationWarning,
                        stacklevel=stacklevel,
                    )
            return h
        if isinstance(headers, Headers):
            return headers
        if headers is None:
            return Headers({})

        warnings.warn(
            (
                "headers must be a dict, twisted.web.http_headers.Headers, or None,"
                " but found {}, which will be ignored."
                " This will raise TypeError in the next treq release."
            ).format(type(headers)),
            DeprecationWarning,
            stacklevel=stacklevel,
        )
        return Headers({})

    def _request_body(self, data, files, json, stacklevel):
        """
        Here we choose a right producer based on the parameters passed in.

        :params data:
            Arbitrary request body data.

            If *files* is also passed this must be a :class:`dict`,
            a :class:`tuple` or :class:`list` of field tuples as accepted by
            :class:`MultiPartProducer`. The request is assigned a Content-Type
            of ``multipart/form-data``.

            If a :class:`dict`, :class:`list`, or :class:`tuple` it is
            URL-encoded and the request assigned a Content-Type of
            ``application/x-www-form-urlencoded``.

            Otherwise, any non-``None`` value is passed to the client's
            *data_to_body_producer* callable (by default,
            :class:`IBodyProducer`), which accepts file-like objects.

        :params files:
            Files to include in the request body, in any of the several formats
            described in :func:`_convert_files()`.

        :params json:
            JSON-encodable data, or the sentinel `_NOTHING`. The sentinel is
            necessary because ``None`` is a valid JSON value.
        """
        if json is not _NOTHING and (files or data):
            warnings.warn(
                (
                    "Argument 'json' will be ignored because '{}' was also passed."
                    " This will raise TypeError in the next treq release."
                ).format("data" if data else "files"),
                DeprecationWarning,
                stacklevel=stacklevel,
            )

        if files:
            # If the files keyword is present we will issue a
            # multipart/form-data request as it suits better for cases
            # with files and/or large objects.
            files = list(_convert_files(files))
            boundary = str(uuid.uuid4()).encode('ascii')
            if data:
                data = _convert_params(data)
            else:
                data = []

            return (
                multipart.MultiPartProducer(data + files, boundary=boundary),
                b'multipart/form-data; boundary=' + boundary,
            )

        # Otherwise stick to x-www-form-urlencoded format
        # as it's generally faster for smaller requests.
        if isinstance(data, (dict, list, tuple)):
            return (
                self._data_to_body_producer(urlencode(data, doseq=True)),
                b'application/x-www-form-urlencoded',
            )
        elif data:
            return (
                self._data_to_body_producer(data),
                None,
            )

        if json is not _NOTHING:
            return (
                self._data_to_body_producer(
                    json_dumps(json, separators=(u',', u':')).encode('utf-8'),
                ),
                b'application/json; charset=UTF-8',
            )

        return None, None


def _convert_params(params):
    if hasattr(params, "iteritems"):
        return list(sorted(params.iteritems()))
    elif hasattr(params, "items"):
        return list(sorted(params.items()))
    elif isinstance(params, (tuple, list)):
        return list(params)
    else:
        raise ValueError("Unsupported format")


def _convert_files(files):
    """Files can be passed in a variety of formats:

        * {'file': open("bla.f")}
        * {'file': (name, open("bla.f"))}
        * {'file': (name, content-type, open("bla.f"))}
        * Anything that has iteritems method, e.g. MultiDict:
          MultiDict([(name, open()), (name, open())]

        Our goal is to standardize it to unified form of:

        * [(param, (file name, content type, producer))]
    """

    if hasattr(files, "iteritems"):
        files = files.iteritems()
    elif hasattr(files, "items"):
        files = files.items()

    for param, val in files:
        file_name, content_type, fobj = (None, None, None)
        if isinstance(val, tuple):
            if len(val) == 2:
                file_name, fobj = val
            elif len(val) == 3:
                file_name, content_type, fobj = val
            else:
                # NB: This is TypeError for backward compatibility. This case
                # used to fall through to `IBodyProducer`, below, which raised
                # TypeError about being unable to coerce None.
                raise TypeError(
                    (
                        "`files` argument must be a sequence of tuples of"
                        " (file_name, file_obj) or"
                        " (file_name, content_type, file_obj),"
                        " but the {!r} tuple has length {}: {!r}"
                    ).format(param, len(val), val),
                )
        else:
            fobj = val
            if hasattr(fobj, "name"):
                file_name = FilePath(fobj.name).basename()

        if not content_type:
            content_type = _guess_content_type(file_name)

        # XXX: Shouldn't this call self._data_to_body_producer?
        yield (param, (file_name, content_type, IBodyProducer(fobj)))


def _query_quote(v):
    # (Any) -> Text
    """
    Percent-encode a querystring name or value.

    :param v: A value.

    :returns:
        The value, coerced to a string and percent-encoded as appropriate for
        a querystring (with space as ``+``).
    """
    if not isinstance(v, (str, bytes)):
        v = str(v)
    if not isinstance(v, bytes):
        v = v.encode("utf-8")
    q = quote_plus(v)
    if isinstance(q, bytes):
        # Python 2.7 returnes bytes when given bytes, but Python 3.x always
        # returns str.  Coverage disabled here to stop Coveralls complaining
        # until we can drop Python 2.7 support.
        q = q.decode("ascii")  # pragma: no cover
    return q


def _coerced_query_params(params):
    """
    Carefully coerce *params* in the same way as `urllib.parse.urlencode()`

    Parameter names and values are coerced to unicode, which is encoded as
    UTF-8 and then percent-encoded. As a special case, `bytes` are directly
    percent-encoded.

    :param params:
        A mapping or sequence of (name, value) two-tuples. The value may be
        a list or tuple of multiple values. Names and values may be pretty much
        any type.

    :returns:
        A generator that yields two-tuples containing percent-encoded text
        strings.
    :rtype:
        Iterator[Tuple[Text, Text]]
    """
    if isinstance(params, Mapping):
        items = params.items()
    else:
        items = params

    for key, values in items:
        key_quoted = _query_quote(key)

        if not isinstance(values, (list, tuple)):
            values = (values,)
        for value in values:
            yield key_quoted, _query_quote(value)


def _from_bytes(orig_bytes):
    return FileBodyProducer(io.BytesIO(orig_bytes))


def _from_file(orig_file):
    return FileBodyProducer(orig_file)


def _guess_content_type(filename):
    if filename:
        guessed = mimetypes.guess_type(filename)[0]
    else:
        guessed = None
    return guessed or 'application/octet-stream'


registerAdapter(_from_bytes, bytes, IBodyProducer)
registerAdapter(_from_file, io.BytesIO, IBodyProducer)

# file()/open() equiv
registerAdapter(_from_file, io.BufferedReader, IBodyProducer)



# Copyright (c) The treq Authors.
# See LICENSE for details.
from typing import Callable, List, Optional, Tuple

import attr
from twisted.internet.defer import Deferred
from twisted.web.http_headers import Headers
from twisted.web.iweb import IAgent, IBodyProducer, IResponse
from zope.interface import implementer


@attr.s(frozen=True, order=False, slots=True)
class RequestRecord:
    """
    The details of a call to :meth:`_AgentSpy.request`

    :ivar method: The *method* argument to :meth:`IAgent.request`
    :ivar uri: The *uri* argument to :meth:`IAgent.request`
    :ivar headers: The *headers* argument to :meth:`IAgent.request`
    :ivar bodyProducer: The *bodyProducer* argument to :meth:`IAgent.request`
    :ivar deferred: The :class:`Deferred` returned by :meth:`IAgent.request`
    """

    method = attr.ib()  # type: bytes
    uri = attr.ib()  # type: bytes
    headers = attr.ib()  # type: Optional[Headers]
    bodyProducer = attr.ib()  # type: Optional[IBodyProducer]
    deferred = attr.ib()  # type: Deferred


@implementer(IAgent)
@attr.s
class _AgentSpy:
    """
    An agent that records HTTP requests

    :ivar _callback:
        A function called with each :class:`RequestRecord`
    """

    _callback = attr.ib()  # type: Callable[Tuple[RequestRecord], None]

    def request(self, method, uri, headers=None, bodyProducer=None):
        # type: (bytes, bytes, Optional[Headers], Optional[IBodyProducer]) -> Deferred[IResponse]  # noqa
        if not isinstance(method, bytes):
            raise TypeError(
                "method must be bytes, not {!r} of type {}".format(method, type(method))
            )
        if not isinstance(uri, bytes):
            raise TypeError(
                "uri must be bytes, not {!r} of type {}".format(uri, type(uri))
            )
        if headers is not None and not isinstance(headers, Headers):
            raise TypeError(
                "headers must be {}, not {!r} of type {}".format(
                    type(Headers), headers, type(headers)
                )
            )
        if bodyProducer is not None and not IBodyProducer.providedBy(bodyProducer):
            raise TypeError(
                (
                    "bodyProducer must implement IBodyProducer, but {!r} does not."
                    " Is the implementation marked with @implementer(IBodyProducer)?"
                ).format(bodyProducer)
            )
        d = Deferred()
        record = RequestRecord(method, uri, headers, bodyProducer, d)
        self._callback(record)
        return d


def agent_spy():
    # type: () -> Tuple[IAgent, List[RequestRecord]]
    """
    Record HTTP requests made with an agent

    This is suitable for low-level testing of wrapper agents. It validates
    the parameters of each call to :meth:`IAgent.request` (synchronously
    raising :exc:`TypeError`) and captures them as a :class:`RequestRecord`,
    which can then be used to inspect the request or generate a response by
    firing the :attr:`~RequestRecord.deferred`.

    :returns:
        A two-tuple of:

         - An :class:`twisted.web.iweb.IAgent`
         - A list of calls made to the agent's
           :meth:`~twisted.web.iweb.IAgent.request()` method
    """
    records = []
    agent = _AgentSpy(records.append)
    return agent, records



from twisted.internet import defer
from twisted.trial.unittest import SynchronousTestCase
from twisted.web import http

from treq.testing import StubTreq, HasHeaders
from treq.testing import RequestSequence, StringStubbingResource


@defer.inlineCallbacks
def make_a_request(treq):
    """
    Make a request using treq.
    """
    response = yield treq.get('http://an.example/foo', params={'a': 'b'},
                              headers={b'Accept': b'application/json'})
    if response.code == http.OK:
        result = yield response.json()
    else:
        message = yield response.text()
        raise Exception("Got an error from the server: {}".format(message))
    defer.returnValue(result)


class MakeARequestTests(SynchronousTestCase):
    """
    Test :func:`make_a_request()` using :mod:`treq.testing.RequestSequence`.
    """

    def test_200_ok(self):
        """On a 200 response, return the response's JSON."""
        req_seq = RequestSequence([
            ((b'get', 'http://an.example/foo', {b'a': [b'b']},
              HasHeaders({'Accept': ['application/json']}), b''),
             (http.OK, {b'Content-Type': b'application/json'}, b'{"status": "ok"}'))
        ])
        treq = StubTreq(StringStubbingResource(req_seq))

        with req_seq.consume(self.fail):
            result = self.successResultOf(make_a_request(treq))

        self.assertEqual({"status": "ok"}, result)

    def test_418_teapot(self):
        """On an unexpected response code, raise an exception"""
        req_seq = RequestSequence([
            ((b'get', 'http://an.example/foo', {b'a': [b'b']},
              HasHeaders({'Accept': ['application/json']}), b''),
             (418, {b'Content-Type': b'text/plain'}, b"I'm a teapot!"))
        ])
        treq = StubTreq(StringStubbingResource(req_seq))

        with req_seq.consume(self.fail):
            failure = self.failureResultOf(make_a_request(treq))

        self.assertEqual(u"Got an error from the server: I'm a teapot!",
                         failure.getErrorMessage())



"""
A local ``httpbin`` server to run integration tests against.

This ensures tests do not depend on `httpbin <https://httpbin.org/>`_.
"""
from __future__ import print_function
import argparse
import datetime
import sys

import httpbin

from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.endpoints import TCP4ServerEndpoint, SSL4ServerEndpoint
from twisted.internet.task import react
from twisted.internet.ssl import (Certificate,
                                  CertificateOptions)

from OpenSSL.crypto import PKey, X509

from twisted.python.threadpool import ThreadPool
from twisted.web.server import Site
from twisted.web.wsgi import WSGIResource

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa

from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives.serialization import Encoding

from .shared import _HTTPBinDescription


def _certificates_for_authority_and_server(service_identity, key_size=2048):
    """
    Create a self-signed CA certificate and server certificate signed
    by the CA.

    :param service_identity: The identity (hostname) of the server.
    :type service_identity: :py:class:`unicode`

    :param key_size: (optional) The size of CA's and server's private
        RSA keys.  Defaults to 2048 bits, which is the minimum allowed
        by OpenSSL Contexts at the default security level.
    :type key_size: :py:class:`int`

    :return: a 3-tuple of ``(certificate_authority_certificate,
             server_private_key, server_certificate)``.
    :rtype: :py:class:`tuple` of (:py:class:`sslverify.Certificate`,
            :py:class:`OpenSSL.crypto.PKey`,
            :py:class:`OpenSSL.crypto.X509`)
    """
    common_name_for_ca = x509.Name(
        [x509.NameAttribute(NameOID.COMMON_NAME, u'Testing Example CA')]
    )
    common_name_for_server = x509.Name(
        [x509.NameAttribute(NameOID.COMMON_NAME, u'Testing Example Server')]
    )
    one_day = datetime.timedelta(1, 0, 0)
    private_key_for_ca = rsa.generate_private_key(
        public_exponent=65537,
        key_size=key_size,
        backend=default_backend()
    )
    public_key_for_ca = private_key_for_ca.public_key()
    ca_certificate = (
        x509.CertificateBuilder()
        .subject_name(common_name_for_ca)
        .issuer_name(common_name_for_ca)
        .not_valid_before(datetime.datetime.today() - one_day)
        .not_valid_after(datetime.datetime.today() + one_day)
        .serial_number(x509.random_serial_number())
        .public_key(public_key_for_ca)
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=9), critical=True,
        )
        .sign(
            private_key=private_key_for_ca, algorithm=hashes.SHA256(),
            backend=default_backend()
        )
    )
    private_key_for_server = rsa.generate_private_key(
        public_exponent=65537,
        key_size=key_size,
        backend=default_backend()
    )
    public_key_for_server = private_key_for_server.public_key()
    server_certificate = (
        x509.CertificateBuilder()
        .subject_name(common_name_for_server)
        .issuer_name(common_name_for_ca)
        .not_valid_before(datetime.datetime.today() - one_day)
        .not_valid_after(datetime.datetime.today() + one_day)
        .serial_number(x509.random_serial_number())
        .public_key(public_key_for_server)
        .add_extension(
            x509.BasicConstraints(ca=False, path_length=None), critical=True,
        )
        .add_extension(
            x509.SubjectAlternativeName(
                [x509.DNSName(service_identity)]
            ),
            critical=True,
        )
        .sign(
            private_key=private_key_for_ca, algorithm=hashes.SHA256(),
            backend=default_backend()
        )
    )

    ca_self_cert = Certificate.loadPEM(
        ca_certificate.public_bytes(Encoding.PEM)
    )

    pkey = PKey.from_cryptography_key(private_key_for_server)
    x509_server_certificate = X509.from_cryptography(server_certificate)

    return ca_self_cert, pkey, x509_server_certificate


def _make_httpbin_site(reactor, threadpool_factory=ThreadPool):
    """
    Return a :py:class:`Site` that hosts an ``httpbin`` WSGI
    application.

    :param reactor: The reactor.
    :param threadpool_factory: (optional) A callable that creates a
        :py:class:`ThreadPool`.

    :return: A :py:class:`Site` that hosts ``httpbin``
    """
    wsgi_threads = threadpool_factory()
    wsgi_threads.start()
    reactor.addSystemEventTrigger("before", "shutdown", wsgi_threads.stop)

    wsgi_resource = WSGIResource(reactor, wsgi_threads, httpbin.app)

    return Site(wsgi_resource)


@inlineCallbacks
def _serve_tls(reactor, host, port, site):
    """
    Serve a site over TLS.

    :param reactor: The reactor.
    :param host: The host on which to listen.
    :type host: :py:class:`str`

    :param port: The host on which to listen.
    :type port: :py:class:`int`
    :type site: The :py:class:`Site` to serve.

    :return: A :py:class:`Deferred` that fires with a
             :py:class:`_HTTPBinDescription`
    """
    (
        ca_cert, private_key, certificate,
    ) = _certificates_for_authority_and_server(host)

    context_factory = CertificateOptions(privateKey=private_key,
                                         certificate=certificate)

    endpoint = SSL4ServerEndpoint(reactor,
                                  port,
                                  sslContextFactory=context_factory,
                                  interface=host)

    port = yield endpoint.listen(site)

    description = _HTTPBinDescription(host=host,
                                      port=port.getHost().port,
                                      cacert=ca_cert.dumpPEM().decode('ascii'))

    return description


@inlineCallbacks
def _serve_tcp(reactor, host, port, site):
    """
    Serve a site over plain TCP.

    :param reactor: The reactor.
    :param host: The host on which to listen.
    :type host: :py:class:`str`

    :param port: The host on which to listen.
    :type port: :py:class:`int`

    :return: A :py:class:`Deferred` that fires with a
             :py:class:`_HTTPBinDescription`
    """
    endpoint = TCP4ServerEndpoint(reactor, port, interface=host)

    port = yield endpoint.listen(site)

    description = _HTTPBinDescription(host=host, port=port.getHost().port)

    return description


def _output_process_description(description, stdout=sys.stdout):
    """
    Write a process description to standard out.

    :param description: The process description.
    :type description: :py:class:`_HTTPBinDescription`

    :param stdout: (optional) Standard out.
    """
    stdout.buffer.write(description.to_json_bytes() + b'\n')
    stdout.buffer.flush()


def _forever_httpbin(reactor, argv,
                     _make_httpbin_site=_make_httpbin_site,
                     _serve_tcp=_serve_tcp,
                     _serve_tls=_serve_tls,
                     _output_process_description=_output_process_description):
    """
    Run ``httpbin`` forever.

    :param reactor: The Twisted reactor.
    :param argv: The arguments with which the script was ran.
    :type argv: :py:class:`list` of :py:class:`str`

    :return: a :py:class:`Deferred` that never fires.
    """
    parser = argparse.ArgumentParser(
        description="""
                    Run httpbin forever.  This writes a JSON object to
                    standard out.  The host and port properties
                    contain the host and port on which httpbin
                    listens.  When run with HTTPS, the cacert property
                    contains the PEM-encode CA certificate that
                    clients must trust.
                    """
    )
    parser.add_argument("--https",
                        help="Serve HTTPS",
                        action="store_const",
                        dest='serve',
                        const=_serve_tls,
                        default=_serve_tcp)
    parser.add_argument("--host",
                        help="The host on which the server will listen.",
                        type=str,
                        default="localhost")
    parser.add_argument("--port",
                        help="The on which the server will listen.",
                        type=int,
                        default=0)

    arguments = parser.parse_args(argv)

    site = _make_httpbin_site(reactor)

    description_deferred = arguments.serve(reactor,
                                           arguments.host,
                                           arguments.port,
                                           site)
    description_deferred.addCallback(_output_process_description)
    description_deferred.addCallback(lambda _: Deferred())

    return description_deferred


if __name__ == '__main__':
    react(_forever_httpbin, (sys.argv[1:],))



# -*- encoding: utf-8 -*-
from hyperlink import DecodedURL
from twisted.internet.task import react
from _utils import print_response

import treq

def main(reactor):
    url = (
        DecodedURL.from_text(u"https://httpbin.org")
        .child(u"get")      # add path /get
        .add(u"foo", u"&")  # add query ?foo=%26
    )
    print(url.to_text())
    return treq.get(url).addCallback(print_response)

react(main, [])



"""
Tests for :py:mod:`treq.test.local_httpbin.shared`
"""
from twisted.trial import unittest

from .. import shared


class HTTPBinDescriptionTests(unittest.SynchronousTestCase):
    """
    Tests for :py:class:`shared._HTTPBinDescription`
    """

    def test_round_trip(self):
        """
        :py:class:`shared._HTTPBinDescription.from_json_bytes` can
        deserialize the output of
        :py:class:`shared._HTTPBinDescription.to_json_bytes`
        """
        original = shared._HTTPBinDescription(host="host", port=123)
        round_tripped = shared._HTTPBinDescription.from_json_bytes(
            original.to_json_bytes(),
        )

        self.assertEqual(original, round_tripped)

    def test_round_trip_cacert(self):
        """
        :py:class:`shared._HTTPBinDescription.from_json_bytes` can
        deserialize the output of
        :py:class:`shared._HTTPBinDescription.to_json_bytes` when
        ``cacert`` is set.
        """
        original = shared._HTTPBinDescription(host="host",
                                              port=123,
                                              cacert='cacert')
        round_tripped = shared._HTTPBinDescription.from_json_bytes(
            original.to_json_bytes(),
        )

        self.assertEqual(original, round_tripped)



"""
In-memory treq returns stubbed responses.
"""
from functools import partial
from inspect import getmembers, isfunction

from unittest.mock import ANY

from twisted.trial.unittest import TestCase
from twisted.web.client import ResponseFailed
from twisted.web.error import SchemeNotSupported
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET

import treq

from treq.testing import (
    HasHeaders,
    RequestSequence,
    StringStubbingResource,
    StubTreq
)


class _StaticTestResource(Resource):
    """Resource that always returns 418 "I'm a teapot"""
    isLeaf = True

    def render(self, request):
        request.setResponseCode(418)
        request.setHeader(b"x-teapot", b"teapot!")
        return b"I'm a teapot"


class _NonResponsiveTestResource(Resource):
    """Resource that returns NOT_DONE_YET and never finishes the request"""
    isLeaf = True

    def render(self, request):
        return NOT_DONE_YET


class _EventuallyResponsiveTestResource(Resource):
    """
    Resource that returns NOT_DONE_YET and stores the request so that something
    else can finish the response later.
    """
    isLeaf = True

    def render(self, request):
        self.stored_request = request
        return NOT_DONE_YET


class _SessionIdTestResource(Resource):
    """
    Resource that returns the current session ID.
    """
    isLeaf = True

    def __init__(self):
        super().__init__()
        # keep track of all sessions created, so we can manually expire them later
        self.sessions = []

    def render(self, request):
        session = request.getSession()
        if session not in self.sessions:
            # new session, add to internal list
            self.sessions.append(session)
        uid = session.uid
        return uid

    def expire_sessions(self):
        """
        Manually expire all sessions created by this resource.
        """
        for session in self.sessions:
            session.expire()
        self.sessions = []


class StubbingTests(TestCase):
    """
    Tests for :class:`StubTreq`.
    """
    def test_stubtreq_provides_all_functions_in_treq_all(self):
        """
        Every single function and attribute exposed by :obj:`treq.__all__` is
        provided by :obj:`StubTreq`.
        """
        treq_things = [(name, obj) for name, obj in getmembers(treq)
                       if name in treq.__all__]
        stub = StubTreq(_StaticTestResource())

        api_things = [(name, obj) for name, obj in treq_things
                      if obj.__module__ == "treq.api"]
        content_things = [(name, obj) for name, obj in treq_things
                          if obj.__module__ == "treq.content"]

        # sanity checks - this test should fail if treq exposes a new API
        # without changes being made to StubTreq and this test.
        msg = ("At the time this test was written, StubTreq only knew about "
               "treq exposing functions from treq.api and treq.content.  If "
               "this has changed, StubTreq will need to be updated, as will "
               "this test.")
        self.assertTrue(all(isfunction(obj) for name, obj in treq_things), msg)
        self.assertEqual(set(treq_things), set(api_things + content_things),
                         msg)

        for name, obj in api_things:
            self.assertTrue(
                isfunction(getattr(stub, name, None)),
                "StubTreq.{0} should be a function.".format(name))

        for name, obj in content_things:
            self.assertIs(
                getattr(stub, name, None), obj,
                "StubTreq.{0} should just expose treq.{0}".format(name))

    def test_providing_resource_to_stub_treq(self):
        """
        The resource provided to StubTreq responds to every request no
        matter what the URI or parameters or data.
        """
        verbs = ('GET', 'PUT', 'HEAD', 'PATCH', 'DELETE', 'POST')
        urls = (
            'http://supports-http.com',
            'https://supports-https.com',
            'http://this/has/a/path/and/invalid/domain/name',
            'https://supports-https.com:8080',
            'http://supports-http.com:8080',
        )
        params = (None, {}, {b'page': [1]})
        headers = (None, {}, {b'x-random-header': [b'value', b'value2']})
        data = (None, b"", b'some data', b'{"some": "json"}')

        stub = StubTreq(_StaticTestResource())

        combos = (
            (verb, {"url": url, "params": p, "headers": h, "data": d})
            for verb in verbs
            for url in urls
            for p in params
            for h in headers
            for d in data
        )
        for combo in combos:
            verb, kwargs = combo
            deferreds = (stub.request(verb, **kwargs),
                         getattr(stub, verb.lower())(**kwargs))
            for d in deferreds:
                resp = self.successResultOf(d)
                self.assertEqual(418, resp.code)
                self.assertEqual([b'teapot!'],
                                 resp.headers.getRawHeaders(b'x-teapot'))
                self.assertEqual(b"" if verb == "HEAD" else b"I'm a teapot",
                                 self.successResultOf(stub.content(resp)))

    def test_handles_invalid_schemes(self):
        """
        Invalid URLs errback with a :obj:`SchemeNotSupported` failure, and does
        so even after a successful request.
        """
        stub = StubTreq(_StaticTestResource())
        self.failureResultOf(stub.get("x-unknown-1:"), SchemeNotSupported)
        self.successResultOf(stub.get("http://url.com"))
        self.failureResultOf(stub.get("x-unknown-2:"), SchemeNotSupported)

    def test_files_are_rejected(self):
        """
        StubTreq does not handle files yet - it should reject requests which
        attempt to pass files.
        """
        stub = StubTreq(_StaticTestResource())
        self.assertRaises(
            AssertionError, stub.request,
            'method', 'http://url', files=b'some file')

    def test_passing_in_strange_data_is_rejected(self):
        """
        StubTreq rejects data that isn't list/dictionary/tuple/bytes/unicode.
        """
        stub = StubTreq(_StaticTestResource())
        self.assertRaises(
            AssertionError, stub.request, 'method', 'http://url',
            data=object())
        self.successResultOf(stub.request('method', 'http://url', data={}))
        self.successResultOf(stub.request('method', 'http://url', data=[]))
        self.successResultOf(stub.request('method', 'http://url', data=()))
        self.successResultOf(
            stub.request('method', 'http://url', data=b""))
        self.successResultOf(
            stub.request('method', 'http://url', data=""))

    def test_handles_failing_asynchronous_requests(self):
        """
        Handle a resource returning NOT_DONE_YET and then canceling the
        request.
        """
        stub = StubTreq(_NonResponsiveTestResource())
        d = stub.request('method', 'http://url', data=b"1234")
        self.assertNoResult(d)
        d.cancel()
        self.failureResultOf(d, ResponseFailed)

    def test_handles_successful_asynchronous_requests(self):
        """
        Handle a resource returning NOT_DONE_YET and then later finishing the
        response.
        """
        rsrc = _EventuallyResponsiveTestResource()
        stub = StubTreq(rsrc)
        d = stub.request('method', 'http://example.com/', data=b"1234")
        self.assertNoResult(d)
        rsrc.stored_request.finish()
        stub.flush()
        resp = self.successResultOf(d)
        self.assertEqual(resp.code, 200)

    def test_handles_successful_asynchronous_requests_with_response_data(self):
        """
        Handle a resource returning NOT_DONE_YET and then sending some data in
        the response.
        """
        rsrc = _EventuallyResponsiveTestResource()
        stub = StubTreq(rsrc)
        d = stub.request('method', 'http://example.com/', data=b"1234")
        self.assertNoResult(d)

        chunks = []
        rsrc.stored_request.write(b'spam ')
        rsrc.stored_request.write(b'eggs')
        stub.flush()
        resp = self.successResultOf(d)
        d = stub.collect(resp, chunks.append)
        self.assertNoResult(d)
        self.assertEqual(b''.join(chunks), b'spam eggs')

        rsrc.stored_request.finish()
        stub.flush()
        self.successResultOf(d)

    def test_handles_successful_asynchronous_requests_with_streaming(self):
        """
        Handle a resource returning NOT_DONE_YET and then streaming data back
        gradually over time.
        """
        rsrc = _EventuallyResponsiveTestResource()
        stub = StubTreq(rsrc)
        d = stub.request('method', 'http://example.com/', data="1234")
        self.assertNoResult(d)

        chunks = []
        rsrc.stored_request.write(b'spam ')
        rsrc.stored_request.write(b'eggs')
        stub.flush()
        resp = self.successResultOf(d)
        d = stub.collect(resp, chunks.append)
        self.assertNoResult(d)
        self.assertEqual(b''.join(chunks), b'spam eggs')

        del chunks[:]
        rsrc.stored_request.write(b'eggs\r\nspam\r\n')
        stub.flush()
        self.assertNoResult(d)
        self.assertEqual(b''.join(chunks), b'eggs\r\nspam\r\n')

        rsrc.stored_request.finish()
        stub.flush()
        self.successResultOf(d)

    def test_session_persistence_between_requests(self):
        """
        Calling request.getSession() in the wrapped resource will return
        a session with the same ID, until the sessions are cleaned.
        """
        rsrc = _SessionIdTestResource()
        stub = StubTreq(rsrc)
        # request 1, getting original session ID
        d = stub.request("method", "http://example.com/")
        resp = self.successResultOf(d)
        cookies = resp.cookies()
        sid_1 = self.successResultOf(resp.content())
        # request 2, ensuring session ID stays the same
        d = stub.request("method", "http://example.com/", cookies=cookies)
        resp = self.successResultOf(d)
        sid_2 = self.successResultOf(resp.content())
        self.assertEqual(sid_1, sid_2)
        # request 3, ensuring the session IDs are different after cleaning
        # or expiring the sessions

        # manually expire the sessions.
        rsrc.expire_sessions()

        d = stub.request("method", "http://example.com/")
        resp = self.successResultOf(d)
        cookies = resp.cookies()
        sid_3 = self.successResultOf(resp.content())
        self.assertNotEqual(sid_1, sid_3)
        # request 4, ensuring that once again the session IDs are the same
        d = stub.request("method", "http://example.com/", cookies=cookies)
        resp = self.successResultOf(d)
        sid_4 = self.successResultOf(resp.content())
        self.assertEqual(sid_3, sid_4)


class HasHeadersTests(TestCase):
    """
    Tests for :obj:`HasHeaders`.
    """
    def test_equality_and_strict_subsets_succeed(self):
        """
        The :obj:`HasHeaders` returns True if both sets of headers are
        equivalent, or the first is a strict subset of the second.
        """
        self.assertEqual(HasHeaders({'one': ['two', 'three']}),
                         {'one': ['two', 'three']},
                         "Equivalent headers do not match.")
        self.assertEqual(HasHeaders({'one': ['two', 'three']}),
                         {'one': ['two', 'three', 'four'],
                          'ten': ['six']},
                         "Strict subset headers do not match")

    def test_partial_or_zero_intersection_subsets_fail(self):
        """
        The :obj:`HasHeaders` returns False if both sets of headers overlap
        but the first is not a strict subset of the second.  It also returns
        False if there is no overlap.
        """
        self.assertNotEqual(HasHeaders({'one': ['two', 'three']}),
                            {'one': ['three', 'four']},
                            "Partial value overlap matches")
        self.assertNotEqual(HasHeaders({'one': ['two', 'three']}),
                            {'one': ['two']},
                            "Missing value matches")
        self.assertNotEqual(HasHeaders({'one': ['two', 'three']}),
                            {'ten': ['six']},
                            "Complete inequality matches")

    def test_case_insensitive_keys(self):
        """
        The :obj:`HasHeaders` equality function ignores the case of the header
        keys.
        """
        self.assertEqual(HasHeaders({b'A': [b'1'], b'b': [b'2']}),
                         {b'a': [b'1'], b'B': [b'2']})

    def test_case_sensitive_values(self):
        """
        The :obj:`HasHeaders` equality function does care about the case of
        the header value.
        """
        self.assertNotEqual(HasHeaders({b'a': [b'a']}), {b'a': [b'A']})

    def test_bytes_encoded_forms(self):
        """
        The :obj:`HasHeaders` equality function compares the bytes-encoded
        forms of both sets of headers.
        """
        self.assertEqual(HasHeaders({b'a': [b'a']}), {u'a': [u'a']})

        self.assertEqual(HasHeaders({u'b': [u'b']}), {b'b': [b'b']})

    def test_repr(self):
        """
        :obj:`HasHeaders` returns a nice string repr.
        """
        self.assertEqual(
            "HasHeaders({b'a': [b'b']})",
            repr(HasHeaders({b"A": [b"b"]})),
        )


class StringStubbingTests(TestCase):
    """
    Tests for :obj:`StringStubbingResource`.
    """
    def _get_response_for(self, expected_args, response):
        """
        Make a :obj:`IStringResponseStubs` that checks the expected args and
        returns the given response.
        """
        method, url, params, headers, data = expected_args

        def get_response_for(_method, _url, _params, _headers, _data):
            self.assertEqual((method, url, params, data),
                             (_method, _url, _params, _data))
            self.assertEqual(HasHeaders(headers), _headers)
            return response

        return get_response_for

    def test_interacts_successfully_with_istub(self):
        """
        The :obj:`IStringResponseStubs` is passed the correct parameters with
        which to evaluate the response, and the response is returned.
        """
        resource = StringStubbingResource(self._get_response_for(
            (b'DELETE', 'http://what/a/thing', {b'page': [b'1']},
             {b'x-header': [b'eh']}, b'datastr'),
            (418, {b'x-response': b'responseheader'}, b'response body')))

        stub = StubTreq(resource)

        d = stub.delete('http://what/a/thing', headers={b'x-header': b'eh'},
                        params={b'page': b'1'}, data=b'datastr')
        resp = self.successResultOf(d)
        self.assertEqual(418, resp.code)
        self.assertEqual([b'responseheader'],
                         resp.headers.getRawHeaders(b'x-response'))
        self.assertEqual(b'response body',
                         self.successResultOf(stub.content(resp)))


class RequestSequenceTests(TestCase):
    """
    Tests for :obj:`RequestSequence`.
    """
    def setUp(self):
        """
        Set up a way to report failures asynchronously.
        """
        self.async_failures = []

    def test_mismatched_request_causes_failure(self):
        """
        If a request is made that is not expected as the next request,
        causes a failure.
        """
        sequence = RequestSequence(
            [((b'get', 'https://anything/', {b'1': [b'2']},
               HasHeaders({b'1': [b'1']}), b'what'),
              (418, {}, b'body')),
             ((b'get', 'http://anything', {},
               HasHeaders({b'2': [b'1']}), b'what'),
              (202, {}, b'deleted'))],
            async_failure_reporter=self.async_failures.append)

        stub = StubTreq(StringStubbingResource(sequence))
        get = partial(stub.get, 'https://anything?1=2', data=b'what',
                      headers={b'1': b'1'})

        resp = self.successResultOf(get())

        self.assertEqual(418, resp.code)
        self.assertEqual(b'body', self.successResultOf(stub.content(resp)))
        self.assertEqual([], self.async_failures)

        resp = self.successResultOf(get())
        self.assertEqual(500, resp.code)
        self.assertEqual(1, len(self.async_failures))
        self.assertIn("Expected the next request to be",
                      self.async_failures[0])

        self.assertFalse(sequence.consumed())

    def test_unexpected_number_of_request_causes_failure(self):
        """
        If there are no more expected requests, making a request causes a
        failure.
        """
        sequence = RequestSequence(
            [],
            async_failure_reporter=self.async_failures.append)
        stub = StubTreq(StringStubbingResource(sequence))
        d = stub.get('https://anything', data=b'what', headers={b'1': b'1'})
        resp = self.successResultOf(d)
        self.assertEqual(500, resp.code)
        self.assertEqual(b'StubbingError',
                         self.successResultOf(resp.content()))
        self.assertEqual(1, len(self.async_failures))
        self.assertIn("No more requests expected, but request",
                      self.async_failures[0])

        # the expected requests have all been made
        self.assertTrue(sequence.consumed())

    def test_works_with_mock_any(self):
        """
        :obj:`mock.ANY` can be used with the request parameters.
        """
        sequence = RequestSequence(
            [((ANY, ANY, ANY, ANY, ANY), (418, {}, b'body'))],
            async_failure_reporter=self.async_failures.append)
        stub = StubTreq(StringStubbingResource(sequence))

        with sequence.consume(sync_failure_reporter=self.fail):
            d = stub.get('https://anything', data=b'what',
                         headers={b'1': b'1'})
            resp = self.successResultOf(d)
            self.assertEqual(418, resp.code)
            self.assertEqual(b'body', self.successResultOf(stub.content(resp)))

        self.assertEqual([], self.async_failures)

        # the expected requests have all been made
        self.assertTrue(sequence.consumed())

    def test_consume_context_manager_fails_on_remaining_requests(self):
        """
        If the `consume` context manager is used, if there are any remaining
        expecting requests, the test case will be failed.
        """
        sequence = RequestSequence(
            [((ANY, ANY, ANY, ANY, ANY), (418, {}, b'body'))] * 2,
            async_failure_reporter=self.async_failures.append)
        stub = StubTreq(StringStubbingResource(sequence))

        consume_failures = []
        with sequence.consume(sync_failure_reporter=consume_failures.append):

            self.successResultOf(stub.get('https://anything', data=b'what',
                                          headers={b'1': b'1'}))

        self.assertEqual(1, len(consume_failures))
        self.assertIn(
            "Not all expected requests were made.  Still expecting:",
            consume_failures[0])
        self.assertIn(
            "{0}(url={0}, params={0}, headers={0}, data={0})".format(
                repr(ANY)),
            consume_failures[0])

        # no asynchronous failures (mismatches, etc.)
        self.assertEqual([], self.async_failures)

    def test_async_failures_logged(self):
        """
        When no `async_failure_reporter` is passed async failures are logged by
        default.
        """
        sequence = RequestSequence([])
        stub = StubTreq(StringStubbingResource(sequence))

        with sequence.consume(self.fail):
            self.successResultOf(stub.get('https://example.com'))

        [failure] = self.flushLoggedErrors()
        self.assertIsInstance(failure.value, AssertionError)



"""
Spawn and monitor an ``httpbin`` child process.
"""
import attr

import signal
import sys
import os


from twisted.protocols import basic, policies
from twisted.internet import protocol, endpoints, error
from twisted.internet.defer import Deferred, succeed

from .shared import _HTTPBinDescription


class _HTTPBinServerProcessProtocol(basic.LineOnlyReceiver):
    """
    Manage the lifecycle of an ``httpbin`` process.
    """
    delimiter = b'\n'

    def __init__(self, all_data_received, terminated):
        """
        Manage the lifecycle of an ``httpbin`` process.

        :param all_data_received: A Deferred that will be called back
            with an :py:class:`_HTTPBinDescription` object
        :type all_data_received: :py:class:`Deferred`

        :param terminated: A Deferred that will be called back when
            the process has ended.
        :type terminated: :py:class:`Deferred`
        """
        self._all_data_received = all_data_received
        self._received = False
        self._terminated = terminated

    def lineReceived(self, line):
        if self._received:
            raise RuntimeError("Unexpected line: {!r}".format(line))
        description = _HTTPBinDescription.from_json_bytes(line)

        self._received = True

        # Remove readers and writers that leave the reactor in a dirty
        # state after a test.
        self.transport.closeStdin()
        self.transport.closeStdout()
        self.transport.closeStderr()

        self._all_data_received.callback(description)

    def connectionLost(self, reason):
        if not self._received:
            self._all_data_received.errback(reason)
        self._terminated.errback(reason)


@attr.s
class _HTTPBinProcess:
    """
    Manage an ``httpbin`` server process.

    :ivar _all_data_received: See
        :py:attr:`_HTTPBinServerProcessProtocol.all_data_received`
    :ivar _terminated: See
        :py:attr:`_HTTPBinServerProcessProtocol.terminated`
    """
    _https = attr.ib()

    _error_log_path = attr.ib(default='httpbin-server-error.log')

    _all_data_received = attr.ib(init=False, default=attr.Factory(Deferred))
    _terminated = attr.ib(init=False, default=attr.Factory(Deferred))

    _process = attr.ib(init=False, default=None)
    _process_description = attr.ib(init=False, default=None)

    _open = staticmethod(open)

    def _spawn_httpbin_process(self, reactor):
        """
        Spawn an ``httpbin`` process, returning a :py:class:`Deferred`
        that fires with the process transport and result.
        """
        server = _HTTPBinServerProcessProtocol(
            all_data_received=self._all_data_received,
            terminated=self._terminated
        )

        argv = [
            sys.executable,
            '-m',
            'treq.test.local_httpbin.child',
        ]

        if self._https:
            argv.append('--https')

        with self._open(self._error_log_path, 'wb') as error_log:
            endpoint = endpoints.ProcessEndpoint(
                reactor,
                sys.executable,
                argv,
                env=os.environ,
                childFDs={
                    1: 'r',
                    2: error_log.fileno(),
                },
            )
            # Processes are spawned synchronously.
            spawned = endpoint.connect(
                # ProtocolWrapper, WrappingFactory's protocol, has a
                # disconnecting attribute.  See
                # https://twistedmatrix.com/trac/ticket/6606
                policies.WrappingFactory(
                    protocol.Factory.forProtocol(lambda: server),
                ),
            )

        def wait_for_protocol(connected_protocol):
            process = connected_protocol.transport
            return self._all_data_received.addCallback(
                return_result_and_process, process,
            )

        def return_result_and_process(description, process):
            return description, process

        return spawned.addCallback(wait_for_protocol)

    def server_description(self, reactor):
        """
        Return a :py:class:`Deferred` that fires with the the process'
        :py:class:`_HTTPBinDescription`, spawning the process if
        necessary.
        """
        if self._process is None:
            ready = self._spawn_httpbin_process(reactor)

            def store_and_schedule_termination(description_and_process):
                description, process = description_and_process

                self._process = process
                self._process_description = description

                reactor.addSystemEventTrigger("before", "shutdown", self.kill)

                return self._process_description

            return ready.addCallback(store_and_schedule_termination)
        else:
            return succeed(self._process_description)

    def kill(self):
        """
        Kill the ``httpbin`` process.
        """
        if not self._process:
            return

        self._process.signalProcess("KILL")

        def suppress_process_terminated(exit_failure):
            exit_failure.trap(error.ProcessTerminated)
            if exit_failure.value.signal != signal.SIGKILL:
                return exit_failure

        return self._terminated.addErrback(suppress_process_terminated)



"""
In-memory version of treq for testing.
"""

from contextlib import contextmanager
from functools import wraps

try:
    from twisted.internet.testing import MemoryReactorClock
except ImportError:
    from twisted.test.proto_helpers import MemoryReactorClock

from twisted.test import iosim

from twisted.internet.address import IPv4Address
from twisted.internet.defer import succeed
from twisted.internet.interfaces import ISSLTransport

from twisted.logger import Logger

from twisted.python.failure import Failure
from twisted.python.urlpath import URLPath

from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.web.client import Agent
from twisted.web.error import SchemeNotSupported
from twisted.web.iweb import IAgent, IAgentEndpointFactory, IBodyProducer
from twisted.web.resource import Resource
from twisted.web.server import Session, Site

from zope.interface import directlyProvides, implementer

import treq
from treq.client import HTTPClient
import attr


@implementer(IAgentEndpointFactory)
@attr.s
class _EndpointFactory:
    """
    An endpoint factory used by :class:`RequestTraversalAgent`.

    :ivar reactor: The agent's reactor.
    :type reactor: :class:`MemoryReactorClock`
    """

    reactor = attr.ib()

    def endpointForURI(self, uri):
        """
        Create an endpoint that represents an in-memory connection to
        a URI.

        Note: This always creates a
        :class:`~twisted.internet.endpoints.TCP4ClientEndpoint` on the
        assumption :class:`RequestTraversalAgent` ignores everything
        about the endpoint but its port.

        :param uri: The URI to connect to.
        :type uri: :class:`~twisted.web.client.URI`

        :return: The endpoint.
        :rtype: An
                :class:`~twisted.internet.interfaces.IStreamClientEndpoint`
                provider.
        """

        if uri.scheme not in {b'http', b'https'}:
            raise SchemeNotSupported("Unsupported scheme: %r" % (uri.scheme,))
        return TCP4ClientEndpoint(self.reactor, "127.0.0.1", uri.port)


@implementer(IAgent)
class RequestTraversalAgent:
    """
    :obj:`~twisted.web.iweb.IAgent` implementation that issues an in-memory
    request rather than going out to a real network socket.
    """

    def __init__(self, rootResource):
        """
        :param rootResource: The Twisted `IResource` at the root of the
            resource tree.
        """
        self._memoryReactor = MemoryReactorClock()
        self._realAgent = Agent.usingEndpointFactory(
            reactor=self._memoryReactor,
            endpointFactory=_EndpointFactory(self._memoryReactor))
        self._rootResource = rootResource
        self._serverFactory = Site(self._rootResource, reactor=self._memoryReactor)
        self._serverFactory.sessionFactory = lambda site, uid: Session(
            site,
            uid,
            reactor=self._memoryReactor,
        )
        self._pumps = set()

    def request(self, method, uri, headers=None, bodyProducer=None):
        """
        Implement IAgent.request.
        """
        # We want to use Agent to parse the HTTP response, so let's ask it to
        # make a request against our in-memory reactor.
        response = self._realAgent.request(method, uri, headers, bodyProducer)

        # If the request has already finished, just propagate the result.  In
        # reality this would only happen in failure, but if the agent ever adds
        # a local cache this might be a success.
        already_called = []

        def check_already_called(r):
            already_called.append(r)
            return r
        response.addBoth(check_already_called)
        if already_called:
            return response

        # That will try to establish an HTTP connection with the reactor's
        # connectTCP method, and MemoryReactor will place Agent's factory into
        # the tcpClients list.  Alternately, it will try to establish an HTTPS
        # connection with the reactor's connectSSL method, and MemoryReactor
        # will place it into the sslClients list.  We'll extract that.
        scheme = URLPath.fromBytes(uri).scheme

        host, port, factory, timeout, bindAddress = (
            self._memoryReactor.tcpClients[-1])

        serverAddress = IPv4Address('TCP', '127.0.0.1', port)
        clientAddress = IPv4Address('TCP', '127.0.0.1', 31337)

        # Create the protocol and fake transport for the client and server,
        # using the factory that was passed to the MemoryReactor for the
        # client, and a Site around our rootResource for the server.
        serverProtocol = self._serverFactory.buildProtocol(clientAddress)
        serverTransport = iosim.FakeTransport(
            serverProtocol, isServer=True,
            hostAddress=serverAddress, peerAddress=clientAddress)
        clientProtocol = factory.buildProtocol(None)
        clientTransport = iosim.FakeTransport(
            clientProtocol, isServer=False,
            hostAddress=clientAddress, peerAddress=serverAddress)

        if scheme == b"https":
            # Provide ISSLTransport on both transports, so everyone knows that
            # this is HTTPS.
            directlyProvides(serverTransport, ISSLTransport)
            directlyProvides(clientTransport, ISSLTransport)

        # Make a pump for wiring the client and server together.
        pump = iosim.connect(
            serverProtocol, serverTransport, clientProtocol, clientTransport)
        self._pumps.add(pump)

        return response

    def flush(self):
        """
        Flush all data between pending client/server pairs.

        This is only necessary if a :obj:`Resource` under test returns
        :obj:`NOT_DONE_YET` from its ``render`` method, making a response
        asynchronous. In that case, after each write from the server,
        :meth:`flush()` must be called so the client can see it.
        """
        old_pumps = self._pumps
        new_pumps = self._pumps = set()
        for p in old_pumps:
            p.flush()
            if p.clientIO.disconnected and p.serverIO.disconnected:
                continue
            new_pumps.add(p)


@implementer(IBodyProducer)
class _SynchronousProducer:
    """
    A partial implementation of an :obj:`IBodyProducer` which produces its
    entire payload immediately.  There is no way to access to an instance of
    this object from :obj:`RequestTraversalAgent` or :obj:`StubTreq`, or even a
    :obj:`Resource: passed to :obj:`StubTreq`.

    This does not implement the :func:`IBodyProducer.stopProducing` method,
    because that is very difficult to trigger.  (The request from
    `RequestTraversalAgent` would have to be canceled while it is still in the
    transmitting state), and the intent is to use `RequestTraversalAgent` to
    make synchronous requests.
    """

    def __init__(self, body):
        """
        Create a synchronous producer with some bytes.
        """
        self.body = body
        msg = ("StubTreq currently only supports url-encodable types, bytes, "
               "or unicode as data.")
        assert isinstance(body, (bytes, str)), msg
        if isinstance(body, str):
            self.body = body.encode('utf-8')
        self.length = len(body)

    def startProducing(self, consumer):
        """
        Immediately produce all data.
        """
        consumer.write(self.body)
        return succeed(None)


def _reject_files(f):
    """
    Decorator that rejects the 'files' keyword argument to the request
    functions, because that is not handled by this yet.
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        if 'files' in kwargs:
            raise AssertionError("StubTreq cannot handle files.")
        return f(*args, **kwargs)
    return wrapper


class StubTreq:
    """
    A fake version of the treq module that can be used for testing that
    provides all the function calls exposed in :obj:`treq.__all__`.
    """
    def __init__(self, resource):
        """
        Construct a client, and pass through client methods and/or
        treq.content functions.

        :param resource: A :obj:`Resource` object that provides the fake
            responses
        """
        self._agent = RequestTraversalAgent(resource)
        _client = HTTPClient(agent=self._agent,
                             data_to_body_producer=_SynchronousProducer)
        for function_name in treq.__all__:
            function = getattr(_client, function_name, None)
            if function is None:
                function = getattr(treq, function_name)
            else:
                function = _reject_files(function)

            setattr(self, function_name, function)
        self.flush = self._agent.flush


class StringStubbingResource(Resource):
    """
    A resource that takes a callable with 5 parameters
    ``(method, url, params, headers, data)`` and returns
    ``(code, headers, body)``.

    The resource uses the callable to return a real response as a result of a
    request.

    The parameters for the callable are:

    - ``method``, the HTTP method as `bytes`.
    - ``url``, the full URL of the request as text.
    - ``params``, a dictionary of query parameters mapping query keys
      lists of values (sorted alphabetically).
    - ``headers``, a dictionary of headers mapping header keys to
      a list of header values (sorted alphabetically).
    - ``data``, the request body as `bytes`.

    The callable must return a ``tuple`` of (code, headers, body) where the
    code is the HTTP status code, the headers is a dictionary of bytes (unlike
    the `headers` parameter, which is a dictionary of lists), and body is
    a string that will be returned as the response body.

    If there is a stubbing error, the return value is undefined (if an
    exception is raised, :obj:`~twisted.web.resource.Resource` will just eat it
    and return 500 in its place).  The callable, or whomever creates the
    callable, should have a way to handle error reporting.
    """
    isLeaf = True

    def __init__(self, get_response_for):
        """
        See :class:`StringStubbingResource`.
        """
        Resource.__init__(self)
        self._get_response_for = get_response_for

    def render(self, request):
        """
        Produce a response according to the stubs provided.
        """
        params = request.args
        headers = {}
        for k, v in request.requestHeaders.getAllRawHeaders():
            headers[k] = v

        for dictionary in (params, headers):
            for k in dictionary:
                dictionary[k] = sorted(dictionary[k])

        # The incoming request does not have the absoluteURI property, because
        # an incoming request is a IRequest, not an IClientRequest, so it
        # the absolute URI needs to be synthesized.

        # But request.URLPath() only returns the scheme and hostname, because
        # that is the URL for this resource (because this resource handles
        # everything from the root on down).

        # So we need to add the request.path (not request.uri, which includes
        # the query parameters)
        absoluteURI = str(request.URLPath().click(request.path))

        status_code, headers, body = self._get_response_for(
            request.method, absoluteURI, params, headers,
            request.content.read())

        request.setResponseCode(status_code)
        for k, v in headers.items():
            request.setHeader(k, v)

        return body


def _maybeEncode(someStr):
    """
    Encode `someStr` to ASCII if required.
    """
    if isinstance(someStr, str):
        return someStr.encode('ascii')
    return someStr


def _maybeEncodeHeaders(headers):
    """ Convert a headers mapping to its bytes-encoded form. """
    return {_maybeEncode(k).lower(): [_maybeEncode(v) for v in vs]
            for k, vs in headers.items()}


class HasHeaders:
    """
    Since Twisted adds headers to a request, such as the host and the content
    length, it's necessary to test whether request headers CONTAIN the expected
    headers (the ones that are not automatically added by Twisted).

    This wraps a set of headers, and can be used in an equality test against
    a superset if the provided headers. The headers keys are lowercased, and
    keys and values are compared in their bytes-encoded forms.

    Headers should be provided as a mapping from strings or bytes to a list of
    strings or bytes.
    """
    def __init__(self, headers):
        self._headers = _maybeEncodeHeaders(headers)

    def __repr__(self):
        return "HasHeaders({0})".format(repr(self._headers))

    def __eq__(self, other_headers):
        compare_to = _maybeEncodeHeaders(other_headers)

        return (set(self._headers.keys()).issubset(set(compare_to.keys())) and
                all([set(v).issubset(set(compare_to[k]))
                     for k, v in self._headers.items()]))

    def __ne__(self, other_headers):
        return not self.__eq__(other_headers)


class RequestSequence:
    """
    For an example usage, see :meth:`RequestSequence.consume`.

    Takes a sequence of::

        [((method, url, params, headers, data), (code, headers, body)),
         ...]

    Expects the requests to arrive in sequence order.  If there are no more
    responses, or the request's parameters do not match the next item's
    expected request parameters, calls `sync_failure_reporter` or
    `async_failure_reporter`.

    For the expected request tuples:

    - ``method`` should be :class:`bytes` normalized to lowercase.
    - ``url`` should be a `str` normalized as per the `transformations in that
      (usually) preserve semantics
      <https://en.wikipedia.org/wiki/URL_normalization>`_.  A URL to
      `http://something-that-looks-like-a-directory` would be normalized to
      `http://something-that-looks-like-a-directory/`
      and a URL to `http://something-that-looks-like-a-page/page.html`
      remains unchanged.
    - ``params`` is a dictionary mapping :class:`bytes` to :class:`list` of
      :class:`bytes`.
    - ``headers`` is a dictionary mapping :class:`bytes` to :class:`list` of
      :class:`bytes` -- note that :class:`twisted.web.client.Agent` may add its
      own headers which are not guaranteed to be present (for instance,
      `user-agent` or `content-length`), so it's better to use some kind of
      matcher like :class:`HasHeaders`.
    - ``data`` is a :class:`bytes`.

    For the response tuples:

    - ``code`` is an integer representing the HTTP status code to return.
    - ``headers`` is a dictionary mapping :class:`bytes` to :class:`bytes` or
      :class:`str`. Note that the value is *not* a list.
    - ``body`` is a :class:`bytes`.

    :ivar list sequence: A sequence of (request tuple, response tuple)
        two-tuples, as described above.
    :ivar async_failure_reporter: An optional callable that takes
        a :class:`str` message indicating a failure. It's asynchronous because
        it cannot just raise an exception—if it does, :meth:`Resource.render
        <twisted.web.resource.Resource.render>` will just convert that into
        a 500 response, and there will be no other failure reporting mechanism.

    When the `async_failure_reporter` parameter is not passed, async failures
    will be reported via a :class:`twisted.logger.Logger` instance, which
    Trial's test case classes (:class:`twisted.trial.unittest.TestCase` and
    :class:`~twisted.trial.unittest.SynchronousTestCase`) will translate into
    a test failure.

    .. note::

        Some versions of
        :class:`twisted.trial.unittest.SynchronousTestCase` report
        logged errors on the wrong test: see `Twisted #9267
        <https://twistedmatrix.com/trac/ticket/9267>`_.

    ..  TODO Update the above note to say what version of
        SynchronousTestCase is fixed once Twisted >17.5.0 is released.

    When not subclassing Trial's classes you must pass `async_failure_reporter`
    and implement equivalent behavior or errors will pass silently. For
    example::

        async_failures = []
        sequence_stubs = RequestSequence([...], async_failures.append)
        stub_treq = StubTreq(StringStubbingResource(sequence_stubs))
        with sequence_stubs.consume(self.fail):  # self = unittest.TestCase
            stub_treq.get('http://fakeurl.com')

        self.assertEqual([], async_failures)
    """
    _log = Logger()

    def __init__(self, sequence, async_failure_reporter=None):
        self._sequence = sequence
        self._async_reporter = async_failure_reporter or self._log_async_error

    def _log_async_error(self, message):
        """
        The default async failure reporter—see `async_failure_reporter`. Logs
        a failure which wraps an :ex:`AssertionError`.

        :param str message: Failure message
        """
        # Passing message twice may look redundant, but Trial only preserves
        # the Failure, not the log message.
        self._log.failure(
            "RequestSequence async error: {message}",
            message=message,
            failure=Failure(AssertionError(message)),
        )

    def consumed(self):
        """
        :return: `bool` representing whether the entire sequence has been
            consumed.  This is useful in tests to assert that the expected
            requests have all been made.
        """
        return len(self._sequence) == 0

    @contextmanager
    def consume(self, sync_failure_reporter):
        """
        Usage::

            sequence_stubs = RequestSequence([...])
            stub_treq = StubTreq(StringStubbingResource(sequence_stubs))
            # self = twisted.trial.unittest.SynchronousTestCase
            with sequence_stubs.consume(self.fail):
                stub_treq.get('http://fakeurl.com')
                stub_treq.get('http://another-fake-url.com')

        If there are still remaining expected requests to be made in the
        sequence, fails the provided test case.

        :param sync_failure_reporter: A callable that takes a single message
            reporting failures.  This can just raise an exception - it does
            not need to be asynchronous, since the exception would not get
            raised within a Resource.

        :return: a context manager that can be used to ensure all expected
            requests have been made.
        """
        yield
        if not self.consumed():
            sync_failure_reporter("\n".join(
                ["Not all expected requests were made.  Still expecting:"] +
                ["- {0}(url={1}, params={2}, headers={3}, data={4})".format(
                    *expected) for expected, _ in self._sequence]))

    def __call__(self, method, url, params, headers, data):
        """
        :return: the next response in the sequence, provided that the
            parameters match the next in the sequence.
        """
        if len(self._sequence) == 0:
            self._async_reporter(
                "No more requests expected, but request {0!r} made.".format(
                    (method, url, params, headers, data)))
            return (500, {}, b"StubbingError")

        expected, response = self._sequence[0]
        e_method, e_url, e_params, e_headers, e_data = expected

        checks = [
            (e_method == method.lower(), "method"),
            (e_url == url, "url"),
            (e_params == params, 'parameters'),
            (e_headers == headers, "headers"),
            (e_data == data, "data")
        ]
        mismatches = [param for success, param in checks if not success]
        if mismatches:
            self._async_reporter(
                "\nExpected the next request to be: {0!r}"
                "\nGot request                    : {1!r}\n"
                "\nMismatches: {2!r}"
                .format(expected, (method, url, params, headers, data),
                        mismatches))
            return (500, {}, b"StubbingError")

        self._sequence = self._sequence[1:]

        return response



from __future__ import absolute_import, division, print_function

from twisted.web.client import Agent, HTTPConnectionPool

from treq.client import HTTPClient


def head(url, **kwargs):
    """
    Make a ``HEAD`` request.

    See :py:func:`treq.request`
    """
    return _client(kwargs).head(url, _stacklevel=4, **kwargs)


def get(url, headers=None, **kwargs):
    """
    Make a ``GET`` request.

    See :py:func:`treq.request`
    """
    return _client(kwargs).get(url, headers=headers, _stacklevel=4, **kwargs)


def post(url, data=None, **kwargs):
    """
    Make a ``POST`` request.

    See :py:func:`treq.request`
    """
    return _client(kwargs).post(url, data=data, _stacklevel=4, **kwargs)


def put(url, data=None, **kwargs):
    """
    Make a ``PUT`` request.

    See :py:func:`treq.request`
    """
    return _client(kwargs).put(url, data=data, _stacklevel=4, **kwargs)


def patch(url, data=None, **kwargs):
    """
    Make a ``PATCH`` request.

    See :py:func:`treq.request`
    """
    return _client(kwargs).patch(url, data=data, _stacklevel=4, **kwargs)


def delete(url, **kwargs):
    """
    Make a ``DELETE`` request.

    See :py:func:`treq.request`
    """
    return _client(kwargs).delete(url, _stacklevel=4, **kwargs)


def request(method, url, **kwargs):
    """
    Make an HTTP request.

    :param str method: HTTP method. Example: ``'GET'``, ``'HEAD'``. ``'PUT'``,
         ``'POST'``.

    :param url: http or https URL, which may include query arguments.
    :type url: :class:`hyperlink.DecodedURL`, `str`, `bytes`, or
        :class:`hyperlink.EncodedURL`

    :param headers: Optional HTTP Headers to send with this request.
    :type headers: Headers or None

    :param params: Optional parameters to be append as the query string to
        the URL, any query string parameters in the URL already will be
        preserved.

    :type params: dict w/ str or list/tuple of str values, list of 2-tuples, or
        None.

    :param data: Optional request body.
    :type data: str, file-like, IBodyProducer, or None

    :param json: Optional JSON-serializable content to pass in body.
    :type json: dict, list/tuple, int, string/unicode, bool, or None

    :param reactor: Optional twisted reactor.

    :param bool persistent: Use persistent HTTP connections.  Default: ``True``
    :param bool allow_redirects: Follow HTTP redirects.  Default: ``True``

    :param auth: HTTP Basic Authentication information --- see
        :func:`treq.auth.add_auth`.
    :type auth: tuple of ``('username', 'password')``.

    :param cookies: Cookies to send with this request.  The HTTP kind, not the
        tasty kind.
    :type cookies: ``dict`` or ``cookielib.CookieJar``

    :param int timeout: Request timeout seconds. If a response is not
        received within this timeframe, a connection is aborted with
        ``CancelledError``.

    :param bool browser_like_redirects: Use browser like redirects
        (i.e. Ignore  RFC2616 section 10.3 and follow redirects from
        POST requests).  Default: ``False``

    :param bool unbuffered: Pass ``True`` to to disable response buffering.  By
        default treq buffers the entire response body in memory.

    :param agent: Provide your own custom agent. Use this to override things
                  like ``connectTimeout`` or ``BrowserLikePolicyForHTTPS``. By
                  default, treq will create its own Agent with reasonable
                  defaults.
    :type agent: twisted.web.iweb.IAgent

    :rtype: Deferred that fires with an IResponse provider.

    .. versionchanged:: treq 20.9.0

        The *url* param now accepts :class:`hyperlink.DecodedURL` and
        :class:`hyperlink.EncodedURL` objects.
    """
    return _client(kwargs).request(method, url, _stacklevel=3, **kwargs)


#
# Private API
#


def default_reactor(reactor):
    """
    Return the specified reactor or the default.
    """
    if reactor is None:
        from twisted.internet import reactor

    return reactor


_global_pool = [None]


def get_global_pool():
    return _global_pool[0]


def set_global_pool(pool):
    _global_pool[0] = pool


def default_pool(reactor, pool, persistent):
    """
    Return the specified pool or a pool with the specified reactor and
    persistence.
    """
    reactor = default_reactor(reactor)

    if pool is not None:
        return pool

    if persistent is False:
        return HTTPConnectionPool(reactor, persistent=persistent)

    if get_global_pool() is None:
        set_global_pool(HTTPConnectionPool(reactor, persistent=True))

    return get_global_pool()


def _client(kwargs):
    agent = kwargs.pop("agent", None)
    pool = kwargs.pop("pool", None)
    persistent = kwargs.pop("persistent", None)
    if agent is None:
        # "reactor" isn't removed from kwargs because it must also be passed
        # down for use in the timeout logic.
        reactor = default_reactor(kwargs.get("reactor"))
        pool = default_pool(reactor, pool, persistent)
        agent = Agent(reactor, pool=pool)
    return HTTPClient(agent)



from collections import OrderedDict
from io import BytesIO

from unittest import mock

from hyperlink import DecodedURL, EncodedURL
from twisted.internet.defer import Deferred, succeed, CancelledError
from twisted.internet.protocol import Protocol
from twisted.python.failure import Failure
from twisted.trial.unittest import TestCase
from twisted.web.client import Agent, ResponseFailed
from twisted.web.http_headers import Headers

from treq.test.util import with_clock
from treq.client import (
    HTTPClient, _BodyBufferingProtocol, _BufferedResponse
)


class HTTPClientTests(TestCase):
    def setUp(self):
        self.agent = mock.Mock(Agent)
        self.client = HTTPClient(self.agent)

        self.fbp_patcher = mock.patch('treq.client.FileBodyProducer')
        self.FileBodyProducer = self.fbp_patcher.start()
        self.addCleanup(self.fbp_patcher.stop)

        self.mbp_patcher = mock.patch('treq.multipart.MultiPartProducer')
        self.MultiPartProducer = self.mbp_patcher.start()
        self.addCleanup(self.mbp_patcher.stop)

    def assertBody(self, expected):
        body = self.FileBodyProducer.mock_calls[0][1][0]
        self.assertEqual(body.read(), expected)

    def test_post(self):
        self.client.post('http://example.com/')
        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_uri_idn(self):
        self.client.request('GET', u'http://č.net')
        self.agent.request.assert_called_once_with(
            b'GET', b'http://xn--bea.net',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_uri_decodedurl(self):
        """
        A URL may be passed as a `hyperlink.DecodedURL` object. It is converted
        to bytes when passed to the underlying agent.
        """
        url = DecodedURL.from_text(u"https://example.org/foo")
        self.client.request("GET", url)
        self.agent.request.assert_called_once_with(
            b"GET", b"https://example.org/foo",
            Headers({b"accept-encoding": [b"gzip"]}),
            None,
        )

    def test_request_uri_encodedurl(self):
        """
        A URL may be passed as a `hyperlink.EncodedURL` object. It is converted
        to bytes when passed to the underlying agent.
        """
        url = EncodedURL.from_text(u"https://example.org/foo")
        self.client.request("GET", url)
        self.agent.request.assert_called_once_with(
            b"GET", b"https://example.org/foo",
            Headers({b"accept-encoding": [b"gzip"]}),
            None,
        )

    def test_request_uri_bytes_pass(self):
        """
        The URL parameter may contain path segments or querystring parameters
        that are not valid UTF-8. These pass through.
        """
        # This URL is http://example.com/hello?who=you, but "hello", "who", and
        # "you" are encoded as UTF-16. The particulars of the encoding aren't
        # important; what matters is that those segments can't be decoded by
        # Hyperlink's UTF-8 default.
        self.client.request(
            "GET",
            (
                "http://example.com/%FF%FEh%00e%00l%00l%00o%00"
                "?%FF%FEw%00h%00o%00=%FF%FEy%00o%00u%00"
            ),
        )
        self.agent.request.assert_called_once_with(
            b'GET',
            (
                b'http://example.com/%FF%FEh%00e%00l%00l%00o%00'
                b'?%FF%FEw%00h%00o%00=%FF%FEy%00o%00u%00'
            ),
            Headers({b'accept-encoding': [b'gzip']}),
            None,
        )

    def test_request_uri_plus_pass(self):
        """
        URL parameters may contain spaces encoded as ``+``. These remain as
        such and are not mangled.

        This reproduces `Klein #339 <https://github.com/twisted/klein/issues/339>`_.
        """
        self.client.request(
            "GET",
            "https://example.com/?foo+bar=baz+biff",
        )
        self.agent.request.assert_called_once_with(
            b'GET',
            b"https://example.com/?foo+bar=baz+biff",
            Headers({b'accept-encoding': [b'gzip']}),
            None,
        )

    def test_request_uri_idn_params(self):
        """
        A URL that contains non-ASCII characters can be augmented with
        querystring parameters.

        This reproduces treq #264.
        """
        self.client.request('GET', u'http://č.net', params={'foo': 'bar'})
        self.agent.request.assert_called_once_with(
            b'GET', b'http://xn--bea.net/?foo=bar',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_uri_hyperlink_params(self):
        """
        The *params* argument augments an instance of `hyperlink.DecodedURL`
        passed as the *url* parameter, just as if it were a string.
        """
        self.client.request(
            method="GET",
            url=DecodedURL.from_text(u"http://č.net"),
            params={"foo": "bar"},
        )
        self.agent.request.assert_called_once_with(
            b"GET", b"http://xn--bea.net/?foo=bar",
            Headers({b"accept-encoding": [b"gzip"]}),
            None,
        )

    def test_request_case_insensitive_methods(self):
        self.client.request('gEt', 'http://example.com/')
        self.agent.request.assert_called_once_with(
            b'GET', b'http://example.com/',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_query_params(self):
        self.client.request('GET', 'http://example.com/',
                            params={'foo': ['bar']})

        self.agent.request.assert_called_once_with(
            b'GET', b'http://example.com/?foo=bar',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_tuple_query_values(self):
        self.client.request('GET', 'http://example.com/',
                            params={'foo': ('bar',)})

        self.agent.request.assert_called_once_with(
            b'GET', b'http://example.com/?foo=bar',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_tuple_query_value_coercion(self):
        """
        treq coerces non-string values passed to *params* like
        `urllib.urlencode()`
        """
        self.client.request('GET', 'http://example.com/', params=[
            ('text', u'A\u03a9'),
            ('text-seq', [u'A\u03a9']),
            ('bytes', [b'ascii']),
            ('bytes-seq', [b'ascii']),
            ('native', ['native']),
            ('native-seq', ['aa', 'bb']),
            ('int', 1),
            ('int-seq', (1, 2, 3)),
            ('none', None),
            ('none-seq', [None, None]),
        ])

        self.agent.request.assert_called_once_with(
            b'GET',
            (
                b'http://example.com/?'
                b'text=A%CE%A9&text-seq=A%CE%A9'
                b'&bytes=ascii&bytes-seq=ascii'
                b'&native=native&native-seq=aa&native-seq=bb'
                b'&int=1&int-seq=1&int-seq=2&int-seq=3'
                b'&none=None&none-seq=None&none-seq=None'
            ),
            Headers({b'accept-encoding': [b'gzip']}),
            None,
        )

    def test_request_tuple_query_param_coercion(self):
        """
        treq coerces non-string param names passed to *params* like
        `urllib.urlencode()`
        """
        # A value used to test that it is never encoded or decoded.
        # It should be invalid UTF-8 or UTF-32 (at least).
        raw_bytes = b"\x00\xff\xfb"

        self.client.request('GET', 'http://example.com/', params=[
            (u'text', u'A\u03a9'),
            (b'bytes', ['ascii', raw_bytes]),
            ('native', 'native'),
            (1, 'int'),
            (None, ['none']),
        ])

        self.agent.request.assert_called_once_with(
            b'GET',
            (
                b'http://example.com/'
                b'?text=A%CE%A9&bytes=ascii&bytes=%00%FF%FB'
                b'&native=native&1=int&None=none'
            ),
            Headers({b'accept-encoding': [b'gzip']}),
            None,
        )

    def test_request_query_param_seps(self):
        """
        When the characters ``&`` and ``#`` are passed to *params* as param
        names or values they are percent-escaped in the URL.

        This reproduces https://github.com/twisted/treq/issues/282
        """
        self.client.request('GET', 'http://example.com/', params=(
            ('ampersand', '&'),
            ('&', 'ampersand'),
            ('octothorpe', '#'),
            ('#', 'octothorpe'),
        ))

        self.agent.request.assert_called_once_with(
            b'GET',
            (
                b'http://example.com/'
                b'?ampersand=%26'
                b'&%26=ampersand'
                b'&octothorpe=%23'
                b'&%23=octothorpe'
            ),
            Headers({b'accept-encoding': [b'gzip']}),
            None,
        )

    def test_request_merge_query_params(self):
        self.client.request('GET', 'http://example.com/?baz=bax',
                            params={'foo': ['bar', 'baz']})

        self.agent.request.assert_called_once_with(
            b'GET', b'http://example.com/?baz=bax&foo=bar&foo=baz',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_merge_tuple_query_params(self):
        self.client.request('GET', 'http://example.com/?baz=bax',
                            params=[('foo', 'bar')])

        self.agent.request.assert_called_once_with(
            b'GET', b'http://example.com/?baz=bax&foo=bar',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_dict_single_value_query_params(self):
        self.client.request('GET', 'http://example.com/',
                            params={'foo': 'bar'})

        self.agent.request.assert_called_once_with(
            b'GET', b'http://example.com/?foo=bar',
            Headers({b'accept-encoding': [b'gzip']}), None)

    def test_request_data_dict(self):
        self.client.request('POST', 'http://example.com/',
                            data={'foo': ['bar', 'baz']})

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/x-www-form-urlencoded'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)

        self.assertBody(b'foo=bar&foo=baz')

    def test_request_data_single_dict(self):
        self.client.request('POST', 'http://example.com/',
                            data={'foo': 'bar'})

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/x-www-form-urlencoded'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)

        self.assertBody(b'foo=bar')

    def test_request_data_tuple(self):
        self.client.request('POST', 'http://example.com/',
                            data=[('foo', 'bar')])

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/x-www-form-urlencoded'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)

        self.assertBody(b'foo=bar')

    def test_request_data_file(self):
        temp_fn = self.mktemp()

        with open(temp_fn, "wb") as temp_file:
            temp_file.write(b'hello')

        self.client.request('POST', 'http://example.com/',
                            data=open(temp_fn, 'rb'))

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)

        self.assertBody(b'hello')

    def test_request_json_dict(self):
        self.client.request('POST', 'http://example.com/', json={'foo': 'bar'})
        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/json; charset=UTF-8'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)
        self.assertBody(b'{"foo":"bar"}')

    def test_request_json_tuple(self):
        self.client.request('POST', 'http://example.com/', json=('foo', 1))
        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/json; charset=UTF-8'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)
        self.assertBody(b'["foo",1]')

    def test_request_json_number(self):
        self.client.request('POST', 'http://example.com/', json=1.)
        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/json; charset=UTF-8'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)
        self.assertBody(b'1.0')

    def test_request_json_string(self):
        self.client.request('POST', 'http://example.com/', json='hello')
        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/json; charset=UTF-8'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)
        self.assertBody(b'"hello"')

    def test_request_json_bool(self):
        self.client.request('POST', 'http://example.com/', json=True)
        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/json; charset=UTF-8'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)
        self.assertBody(b'true')

    def test_request_json_none(self):
        self.client.request('POST', 'http://example.com/', json=None)
        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({b'Content-Type': [b'application/json; charset=UTF-8'],
                     b'accept-encoding': [b'gzip']}),
            self.FileBodyProducer.return_value)
        self.assertBody(b'null')

    @mock.patch('treq.client.uuid.uuid4', mock.Mock(return_value="heyDavid"))
    def test_request_no_name_attachment(self):

        self.client.request(
            'POST', 'http://example.com/', files={"name": BytesIO(b"hello")})

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({
                b'accept-encoding': [b'gzip'],
                b'Content-Type': [b'multipart/form-data; boundary=heyDavid']}),
            self.MultiPartProducer.return_value)

        FP = self.FileBodyProducer.return_value
        self.assertEqual(
            mock.call(
                [('name', (None, 'application/octet-stream', FP))],
                boundary=b'heyDavid'),
            self.MultiPartProducer.call_args)

    @mock.patch('treq.client.uuid.uuid4', mock.Mock(return_value="heyDavid"))
    def test_request_named_attachment(self):

        self.client.request(
            'POST', 'http://example.com/', files={
                "name": ('image.jpg', BytesIO(b"hello"))})

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({
                b'accept-encoding': [b'gzip'],
                b'Content-Type': [b'multipart/form-data; boundary=heyDavid']}),
            self.MultiPartProducer.return_value)

        FP = self.FileBodyProducer.return_value
        self.assertEqual(
            mock.call(
                [('name', ('image.jpg', 'image/jpeg', FP))],
                boundary=b'heyDavid'),
            self.MultiPartProducer.call_args)

    @mock.patch('treq.client.uuid.uuid4', mock.Mock(return_value="heyDavid"))
    def test_request_named_attachment_and_ctype(self):

        self.client.request(
            'POST', 'http://example.com/', files={
                "name": ('image.jpg', 'text/plain', BytesIO(b"hello"))})

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({
                b'accept-encoding': [b'gzip'],
                b'Content-Type': [b'multipart/form-data; boundary=heyDavid']}),
            self.MultiPartProducer.return_value)

        FP = self.FileBodyProducer.return_value
        self.assertEqual(
            mock.call(
                [('name', ('image.jpg', 'text/plain', FP))],
                boundary=b'heyDavid'),
            self.MultiPartProducer.call_args)

    def test_request_files_tuple_too_short(self):
        """
        The `HTTPClient.request()` *files* argument requires tuples of length
        2 or 3. It raises `TypeError` when the tuple is too short.
        """
        with self.assertRaises(TypeError) as c:
            self.client.request(
                "POST",
                b"http://example.com/",
                files=[("t1", ("foo.txt",))],
            )

        self.assertIn("'t1' tuple has length 1", str(c.exception))

    def test_request_files_tuple_too_long(self):
        """
        The `HTTPClient.request()` *files* argument requires tuples of length
        2 or 3. It raises `TypeError` when the tuple is too long.
        """
        with self.assertRaises(TypeError) as c:
            self.client.request(
                "POST",
                b"http://example.com/",
                files=[
                    ("t4", ("foo.txt", "text/plain", BytesIO(b"...\n"), "extra!")),
                ],
            )

        self.assertIn("'t4' tuple has length 4", str(c.exception))

    @mock.patch('treq.client.uuid.uuid4', mock.Mock(return_value="heyDavid"))
    def test_request_mixed_params(self):

        class NamedFile(BytesIO):
            def __init__(self, val):
                BytesIO.__init__(self, val)
                self.name = "image.png"

        self.client.request(
            'POST', 'http://example.com/',
            data=[("a", "b"), ("key", "val")],
            files=[
                ("file1", ('image.jpg', BytesIO(b"hello"))),
                ("file2", NamedFile(b"yo"))])

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({
                b'accept-encoding': [b'gzip'],
                b'Content-Type': [b'multipart/form-data; boundary=heyDavid']}),
            self.MultiPartProducer.return_value)

        FP = self.FileBodyProducer.return_value
        self.assertEqual(
            mock.call([
                ('a', 'b'),
                ('key', 'val'),
                ('file1', ('image.jpg', 'image/jpeg', FP)),
                ('file2', ('image.png', 'image/png', FP))],
                boundary=b'heyDavid'),
            self.MultiPartProducer.call_args)

    @mock.patch('treq.client.uuid.uuid4', mock.Mock(return_value="heyDavid"))
    def test_request_mixed_params_dict(self):

        self.client.request(
            'POST', 'http://example.com/',
            data={"key": "a", "key2": "b"},
            files={"file1": BytesIO(b"hey")})

        self.agent.request.assert_called_once_with(
            b'POST', b'http://example.com/',
            Headers({
                b'accept-encoding': [b'gzip'],
                b'Content-Type': [b'multipart/form-data; boundary=heyDavid']}),
            self.MultiPartProducer.return_value)

        FP = self.FileBodyProducer.return_value
        self.assertEqual(
            mock.call([
                ('key', 'a'),
                ('key2', 'b'),
                ('file1', (None, 'application/octet-stream', FP))],
                boundary=b'heyDavid'),
            self.MultiPartProducer.call_args)

    def test_request_unsupported_params_combination(self):
        self.assertRaises(ValueError,
                          self.client.request,
                          'POST', 'http://example.com/',
                          data=BytesIO(b"yo"),
                          files={"file1": BytesIO(b"hey")})

    def test_request_json_with_data(self):
        """
        Passing `HTTPClient.request()` both *data* and *json* parameters is
        invalid because *json* is ignored. This behavior is deprecated.
        """
        self.client.request(
            "POST",
            "http://example.com/",
            data=BytesIO(b"..."),
            json=None,  # NB: None is a valid value. It encodes to b'null'.
        )

        [w] = self.flushWarnings([self.test_request_json_with_data])
        self.assertEqual(DeprecationWarning, w["category"])
        self.assertEqual(
            (
                "Argument 'json' will be ignored because 'data' was also passed."
                " This will raise TypeError in the next treq release."
            ),
            w['message'],
        )

    def test_request_json_with_files(self):
        """
        Passing `HTTPClient.request()` both *files* and *json* parameters is
        invalid because *json* is ignored. This behavior is deprecated.
        """
        self.client.request(
            "POST",
            "http://example.com/",
            files={"f1": ("foo.txt", "text/plain", BytesIO(b"...\n"))},
            json=["this is ignored"],
        )

        [w] = self.flushWarnings([self.test_request_json_with_files])
        self.assertEqual(DeprecationWarning, w["category"])
        self.assertEqual(
            (
                "Argument 'json' will be ignored because 'files' was also passed."
                " This will raise TypeError in the next treq release."
            ),
            w['message'],
        )

    def test_request_dict_headers(self):
        self.client.request('GET', 'http://example.com/', headers={
            'User-Agent': 'treq/0.1dev',
            'Accept': ['application/json', 'text/plain']
        })

        self.agent.request.assert_called_once_with(
            b'GET', b'http://example.com/',
            Headers({b'User-Agent': [b'treq/0.1dev'],
                     b'accept-encoding': [b'gzip'],
                     b'Accept': [b'application/json', b'text/plain']}),
            None)

    def test_request_headers_object(self):
        """
        The *headers* parameter accepts a `twisted.web.http_headers.Headers`
        instance.
        """
        self.client.request(
            "GET",
            "https://example.com",
            headers=Headers({"X-Foo": ["bar"]}),
        )

        self.agent.request.assert_called_once_with(
            b"GET",
            b"https://example.com",
            Headers({
                "X-Foo": ["bar"],
                "Accept-Encoding": ["gzip"],
            }),
            None,
        )

    def test_request_headers_invalid_type(self):
        """
        `HTTPClient.request()` warns that headers of an unexpected type are
        invalid and that this behavior is deprecated.
        """
        self.client.request('GET', 'http://example.com', headers=[])

        [w] = self.flushWarnings([self.test_request_headers_invalid_type])
        self.assertEqual(DeprecationWarning, w['category'])
        self.assertIn(
            "headers must be a dict, twisted.web.http_headers.Headers, or None,",
            w['message'],
        )

    def test_request_dict_headers_invalid_values(self):
        """
        `HTTPClient.request()` warns that non-string header values are dropped
        and that this behavior is deprecated.
        """
        self.client.request('GET', 'http://example.com', headers=OrderedDict([
            ('none', None),
            ('one', 1),
            ('ok', 'string'),
        ]))

        [w1, w2] = self.flushWarnings([self.test_request_dict_headers_invalid_values])
        self.assertEqual(DeprecationWarning, w1['category'])
        self.assertEqual(DeprecationWarning, w2['category'])
        self.assertIn(
            "The value of headers key 'none' has non-string type",
            w1['message'],
        )
        self.assertIn(
            "The value of headers key 'one' has non-string type",
            w2['message'],
        )

    def test_request_invalid_param(self):
        """
        `HTTPClient.request()` rejects invalid keyword parameters with
        `TypeError`.
        """
        self.assertRaises(
            TypeError,
            self.client.request,
            "GET",
            b"http://example.com",
            invalid=True,
        )

    @with_clock
    def test_request_timeout_fired(self, clock):
        """
        Verify the request is cancelled if a response is not received
        within specified timeout period.
        """
        self.agent.request.return_value = d = Deferred()
        self.client.request('GET', 'http://example.com', timeout=2)

        # simulate we haven't gotten a response within timeout seconds
        clock.advance(3)

        # a deferred should have been cancelled
        self.failureResultOf(d, CancelledError)

    @with_clock
    def test_request_timeout_cancelled(self, clock):
        """
        Verify timeout is cancelled if a response is received before
        timeout period elapses.
        """
        self.agent.request.return_value = d = Deferred()
        self.client.request('GET', 'http://example.com', timeout=2)

        # simulate a response
        d.callback(mock.Mock(code=200, headers=Headers({})))

        # now advance the clock but since we already got a result,
        # a cancellation timer should have been cancelled
        clock.advance(3)

        self.successResultOf(d)

    def test_response_is_buffered(self):
        response = mock.Mock(deliverBody=mock.Mock(),
                             headers=Headers({}))

        self.agent.request.return_value = succeed(response)

        d = self.client.get('http://www.example.com')

        result = self.successResultOf(d)

        protocol = mock.Mock(Protocol)
        result.deliverBody(protocol)
        self.assertEqual(response.deliverBody.call_count, 1)

        result.deliverBody(protocol)
        self.assertEqual(response.deliverBody.call_count, 1)

    def test_response_buffering_is_disabled_with_unbufferred_arg(self):
        response = mock.Mock(headers=Headers({}))

        self.agent.request.return_value = succeed(response)

        d = self.client.get('http://www.example.com', unbuffered=True)

        # YOLO public attribute.
        self.assertEqual(self.successResultOf(d).original, response)

    def test_request_post_redirect_denied(self):
        response = mock.Mock(code=302, headers=Headers({'Location': ['/']}))
        self.agent.request.return_value = succeed(response)
        d = self.client.post('http://www.example.com')
        self.failureResultOf(d, ResponseFailed)

    def test_request_browser_like_redirects(self):
        response = mock.Mock(code=302, headers=Headers({'Location': ['/']}))

        self.agent.request.return_value = succeed(response)

        raw = mock.Mock(return_value=[])
        final_resp = mock.Mock(code=200, headers=mock.Mock(getRawHeaders=raw))
        with mock.patch('twisted.web.client.RedirectAgent._handleRedirect',
                        return_value=final_resp):
            d = self.client.post('http://www.google.com',
                                 browser_like_redirects=True,
                                 unbuffered=True)

        self.assertEqual(self.successResultOf(d).original, final_resp)


class BodyBufferingProtocolTests(TestCase):
    def test_buffers_data(self):
        buffer = []
        protocol = _BodyBufferingProtocol(
            mock.Mock(Protocol),
            buffer,
            None
        )

        protocol.dataReceived("foo")
        self.assertEqual(buffer, ["foo"])

        protocol.dataReceived("bar")
        self.assertEqual(buffer, ["foo", "bar"])

    def test_propagates_data_to_destination(self):
        destination = mock.Mock(Protocol)
        protocol = _BodyBufferingProtocol(
            destination,
            [],
            None
        )

        protocol.dataReceived(b"foo")
        destination.dataReceived.assert_called_once_with(b"foo")

        protocol.dataReceived(b"bar")
        destination.dataReceived.assert_called_with(b"bar")

    def test_fires_finished_deferred(self):
        finished = Deferred()
        protocol = _BodyBufferingProtocol(
            mock.Mock(Protocol),
            [],
            finished
        )

        class TestResponseDone(Exception):
            pass

        protocol.connectionLost(TestResponseDone())

        self.failureResultOf(finished, TestResponseDone)

    def test_propogates_connectionLost_reason(self):
        destination = mock.Mock(Protocol)
        protocol = _BodyBufferingProtocol(
            destination,
            [],
            Deferred().addErrback(lambda ign: None)
        )

        class TestResponseDone(Exception):
            pass

        reason = TestResponseDone()
        protocol.connectionLost(reason)
        destination.connectionLost.assert_called_once_with(reason)


class BufferedResponseTests(TestCase):
    def test_wraps_protocol(self):
        wrappers = []
        wrapped = mock.Mock(Protocol)
        response = mock.Mock(deliverBody=mock.Mock(wraps=wrappers.append))

        br = _BufferedResponse(response)

        br.deliverBody(wrapped)
        response.deliverBody.assert_called_once_with(wrappers[0])
        self.assertNotEqual(wrapped, wrappers[0])

    def test_concurrent_receivers(self):
        wrappers = []
        wrapped = mock.Mock(Protocol)
        unwrapped = mock.Mock(Protocol)
        response = mock.Mock(deliverBody=mock.Mock(wraps=wrappers.append))

        br = _BufferedResponse(response)

        br.deliverBody(wrapped)
        br.deliverBody(unwrapped)
        response.deliverBody.assert_called_once_with(wrappers[0])

        wrappers[0].dataReceived(b"foo")
        wrapped.dataReceived.assert_called_once_with(b"foo")

        self.assertEqual(unwrapped.dataReceived.call_count, 0)

        class TestResponseDone(Exception):
            pass

        done = Failure(TestResponseDone())

        wrappers[0].connectionLost(done)
        wrapped.connectionLost.assert_called_once_with(done)
        unwrapped.dataReceived.assert_called_once_with(b"foo")
        unwrapped.connectionLost.assert_called_once_with(done)

    def test_receiver_after_finished(self):
        wrappers = []
        finished = mock.Mock(Protocol)

        response = mock.Mock(deliverBody=mock.Mock(wraps=wrappers.append))

        br = _BufferedResponse(response)
        br.deliverBody(mock.Mock(Protocol))
        wrappers[0].dataReceived(b"foo")

        class TestResponseDone(Exception):
            pass

        done = Failure(TestResponseDone())

        wrappers[0].connectionLost(done)

        br.deliverBody(finished)

        finished.dataReceived.assert_called_once_with(b"foo")
        finished.connectionLost.assert_called_once_with(done)



# Copyright (c) The treq Authors.
# See LICENSE for details.
from io import BytesIO

from twisted.trial.unittest import SynchronousTestCase
from twisted.web.client import FileBodyProducer
from twisted.web.http_headers import Headers
from twisted.web.iweb import IAgent

from treq._agentspy import RequestRecord, agent_spy


class APISpyTests(SynchronousTestCase):
    """
    The agent_spy API provides an agent that records each request made to it.
    """

    def test_provides_iagent(self):
        """
        The agent returned by agent_spy() provides the IAgent interface.
        """
        agent, _ = agent_spy()

        self.assertTrue(IAgent.providedBy(agent))

    def test_records(self):
        """
        Each request made with the agent is recorded.
        """
        agent, requests = agent_spy()

        body = FileBodyProducer(BytesIO(b"..."))
        d1 = agent.request(b"GET", b"https://foo")
        d2 = agent.request(b"POST", b"http://bar", Headers({}))
        d3 = agent.request(b"PUT", b"https://baz", None, bodyProducer=body)

        self.assertEqual(
            requests,
            [
                RequestRecord(b"GET", b"https://foo", None, None, d1),
                RequestRecord(b"POST", b"http://bar", Headers({}), None, d2),
                RequestRecord(b"PUT", b"https://baz", None, body, d3),
            ],
        )

    def test_record_attributes(self):
        """
        Each parameter passed to `request` is available as an attribute of the
        RequestRecord. Additionally, the deferred returned by the call is
        available.
        """
        agent, requests = agent_spy()
        headers = Headers()
        body = FileBodyProducer(BytesIO(b"..."))

        deferred = agent.request(b"method", b"uri", headers=headers, bodyProducer=body)

        [rr] = requests
        self.assertIs(rr.method, b"method")
        self.assertIs(rr.uri, b"uri")
        self.assertIs(rr.headers, headers)
        self.assertIs(rr.bodyProducer, body)
        self.assertIs(rr.deferred, deferred)

    def test_type_validation(self):
        """
        The request method enforces correctness by raising TypeError when
        passed parameters of the wrong type.
        """
        agent, _ = agent_spy()

        self.assertRaises(TypeError, agent.request, u"method not bytes", b"uri")
        self.assertRaises(TypeError, agent.request, b"method", u"uri not bytes")
        self.assertRaises(
            TypeError, agent.request, b"method", b"uri", {"not": "headers"}
        )
        self.assertRaises(
            TypeError, agent.request, b"method", b"uri", None, b"not ibodyproducer"
        )



from twisted.internet.task import react
from _utils import print_response

import treq


def main(reactor, *args):
    d = treq.get('https://httpbin.org/cookies/set?hello=world')

    def _get_jar(resp):
        jar = resp.cookies()

        print('The server set our hello cookie to: {}'.format(jar['hello']))

        return treq.get('https://httpbin.org/cookies', cookies=jar)

    d.addCallback(_get_jar)
    d.addCallback(print_response)

    return d

react(main, [])



"""
Things shared between the ``httpbin`` child and parent processes
"""
import attr
import json


@attr.s
class _HTTPBinDescription:
    """
    Describe an ``httpbin`` process.

    :param host: The host on which the process listens.
    :type host: :py:class:`str`

    :param port: The port on which the process listens.
    :type port: :py:class:`int`

    :param cacert: (optional) The PEM-encoded certificate authority's
        certificate.  The calling process' treq must trust this when
        running HTTPS tests.
    :type cacert: :py:class:`bytes` or :py:class:`None`
    """
    host = attr.ib()
    port = attr.ib()
    cacert = attr.ib(default=None)

    @classmethod
    def from_json_bytes(cls, json_data):
        """
        Deserialize an instance from JSON bytes.
        """
        return cls(**json.loads(json_data.decode('ascii')))

    def to_json_bytes(self):
        """
        Serialize an instance from JSON bytes.
        """
        return json.dumps(attr.asdict(self), sort_keys=True).encode('ascii')



from twisted.internet.task import react
from twisted.internet.defer import inlineCallbacks

import treq


@inlineCallbacks
def main(reactor):
    print('List of tuples')
    resp = yield treq.get('https://httpbin.org/get',
                          params=[('foo', 'bar'), ('baz', 'bax')])
    content = yield resp.text()
    print(content)

    print('Single value dictionary')
    resp = yield treq.get('https://httpbin.org/get',
                          params={'foo': 'bar', 'baz': 'bax'})
    content = yield resp.text()
    print(content)

    print('Multi value dictionary')
    resp = yield treq.get('https://httpbin.org/get',
                          params={b'foo': [b'bar', b'baz', b'bax']})
    content = yield resp.text()
    print(content)

    print('Mixed value dictionary')
    resp = yield treq.get('https://httpbin.org/get',
                          params={'foo': [1, 2, 3], 'bax': b'quux', b'bar': 'foo'})
    content = yield resp.text()
    print(content)

    print('Preserved query parameters')
    resp = yield treq.get('https://httpbin.org/get?foo=bar',
                          params={'baz': 'bax'})
    content = yield resp.text()
    print(content)

react(main, [])



from setuptools import find_packages, setup

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Framework :: Twisted",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
]

if __name__ == "__main__":

    with open('README.rst') as f:
        readme = f.read()

    setup(
        name="treq",
        packages=find_packages('src'),
        package_dir={"": "src"},
        setup_requires=["incremental"],
        use_incremental=True,
        python_requires='>=3.6',
        install_requires=[
            "incremental",
            "requests >= 2.1.0",
            "hyperlink >= 21.0.0",
            "Twisted[tls] >= 18.7.0",
            "attrs",
        ],
        extras_require={
            "dev": [
                "pep8",
                "pyflakes",
                "httpbin==0.5.0",
            ],
            "docs": [
                "sphinx>=1.4.8",
            ],
        },
        package_data={"treq": ["_version"]},
        author="David Reid",
        author_email="dreid@dreid.org",
        maintainer="Tom Most",
        maintainer_email="twm@freecog.net",
        classifiers=classifiers,
        description="High-level Twisted HTTP Client API",
        license="MIT/X",
        url="https://github.com/twisted/treq",
        long_description=readme,
        long_description_content_type='text/x-rst',
    )



# Copyright 2012-2020 The treq Authors.
# See LICENSE for details.
from __future__ import absolute_import, division, print_function

import binascii
from typing import Union

from twisted.web.http_headers import Headers
from twisted.web.iweb import IAgent
from zope.interface import implementer


class UnknownAuthConfig(Exception):
    """
    The authentication config provided couldn't be interpreted.
    """
    def __init__(self, config):
        super(Exception, self).__init__(
            '{0!r} not of a known type.'.format(config))


@implementer(IAgent)
class _RequestHeaderSetterAgent:
    """
    Wrap an agent to set request headers

    :ivar _agent: The wrapped agent.

    :ivar _request_headers:
        Headers to set on each request before forwarding it to the wrapped
        agent.
    """
    def __init__(self, agent, headers):
        self._agent = agent
        self._headers = headers

    def request(self, method, uri, headers=None, bodyProducer=None):
        if headers is None:
            requestHeaders = self._headers
        else:
            requestHeaders = headers.copy()
            for header, values in self._headers.getAllRawHeaders():
                requestHeaders.setRawHeaders(header, values)

        return self._agent.request(
            method, uri, headers=requestHeaders, bodyProducer=bodyProducer)


def add_basic_auth(agent, username, password):
    # type: (IAgent, Union[str, bytes], Union[str, bytes]) -> IAgent
    """
    Wrap an agent to add HTTP basic authentication

    The returned agent sets the *Authorization* request header according to the
    basic authentication scheme described in :rfc:`7617`. This header contains
    the given *username* and *password* in plaintext, and thus should only be
    used over an encrypted transport (HTTPS).

    Note that the colon (``:``) is used as a delimiter between the *username*
    and *password*, so if either parameter includes a colon the interpretation
    of the *Authorization* header is server-defined.

    :param agent: Agent to wrap.
    :param username: The username.
    :param password: The password.

    :returns: :class:`~twisted.web.iweb.IAgent`
    """
    if not isinstance(username, bytes):
        username = username.encode('utf-8')
    if not isinstance(password, bytes):
        password = password.encode('utf-8')

    creds = binascii.b2a_base64(b'%s:%s' % (username, password)).rstrip(b'\n')
    return _RequestHeaderSetterAgent(
        agent,
        Headers({b'Authorization': [b'Basic ' + creds]}),
    )


def add_auth(agent, auth_config):
    """
    Wrap an agent to perform authentication

    :param agent: Agent to wrap.

    :param auth_config:
        A ``('username', 'password')`` tuple --- see :func:`add_basic_auth`.

    :returns: :class:`~twisted.web.iweb.IAgent`

    :raises UnknownAuthConfig:
        When the format *auth_config* isn't supported.
    """
    if isinstance(auth_config, tuple):
        return add_basic_auth(agent, auth_config[0], auth_config[1])

    raise UnknownAuthConfig(auth_config)



"""
Tests for :py:mod:`treq.test.local_httpbin.parent`
"""
import attr

import json

import signal

import sys

from twisted.internet import defer
from twisted.internet.interfaces import (IProcessTransport,
                                         IReactorCore,
                                         IReactorProcess)
from twisted.python.failure import Failure

from treq.test.util import skip_on_windows_because_of_199

from twisted.internet.error import ProcessTerminated, ConnectionDone

try:
    from twisted.internet.testing import MemoryReactor, StringTransport
except ImportError:
    from twisted.test.proto_helpers import MemoryReactor, StringTransport

from twisted.trial.unittest import SynchronousTestCase

from zope.interface import implementer, verify

from .. import parent, shared


skip = skip_on_windows_because_of_199()


@attr.s
class FakeProcessTransportState(object):
    """
    State for :py:class:`FakeProcessTransport`.
    """
    standard_in_closed = attr.ib(default=False)
    standard_out_closed = attr.ib(default=False)
    standard_error_closed = attr.ib(default=False)
    signals = attr.ib(default=attr.Factory(list))


@implementer(IProcessTransport)
@attr.s
class FakeProcessTransport(StringTransport, object):
    """
    A fake process transport.
    """
    pid = 1234

    _state = attr.ib()

    def closeStdin(self):
        """
        Close standard in.
        """
        self._state.standard_in_closed = True

    def closeStdout(self):
        """
        Close standard out.
        """
        self._state.standard_out_closed = True

    def closeStderr(self):
        """
        Close standard error.
        """
        self._state.standard_error_closed = True

    def closeChildFD(self, descriptor):
        """
        Close a child's file descriptor.

        :param descriptor: See
            :py:class:`IProcessProtocol.closeChildFD`
        """

    def writeToChild(self, childFD, data):
        """
        Write data to a child's file descriptor.

        :param childFD: See :py:class:`IProcessProtocol.writeToChild`
        :param data: See :py:class:`IProcessProtocol.writeToChild`
        """

    def signalProcess(self, signalID):
        """
        Send a signal.

        :param signalID: See
            :py:class:`IProcessProtocol.signalProcess`
        """
        self._state.signals.append(signalID)


class FakeProcessTransportTests(SynchronousTestCase):
    """
    Tests for :py:class:`FakeProcessTransport`.
    """

    def setUp(self):
        self.state = FakeProcessTransportState()
        self.transport = FakeProcessTransport(self.state)

    def test_provides_interface(self):
        """
        Instances provide :py:class:`IProcessTransport`.
        """
        verify.verifyObject(IProcessTransport, self.transport)

    def test_closeStdin(self):
        """
        Closing standard in updates the state instance.
        """
        self.assertFalse(self.state.standard_in_closed)
        self.transport.closeStdin()
        self.assertTrue(self.state.standard_in_closed)

    def test_closeStdout(self):
        """
        Closing standard out updates the state instance.
        """
        self.assertFalse(self.state.standard_out_closed)
        self.transport.closeStdout()
        self.assertTrue(self.state.standard_out_closed)

    def test_closeStderr(self):
        """
        Closing standard error updates the state instance.
        """
        self.assertFalse(self.state.standard_error_closed)
        self.transport.closeStderr()
        self.assertTrue(self.state.standard_error_closed)


class HTTPServerProcessProtocolTests(SynchronousTestCase):
    """
    Tests for :py:class:`parent._HTTPBinServerProcessProtocol`
    """

    def setUp(self):
        self.transport_state = FakeProcessTransportState()
        self.transport = FakeProcessTransport(self.transport_state)

        self.all_data_received = defer.Deferred()
        self.terminated = defer.Deferred()

        self.protocol = parent._HTTPBinServerProcessProtocol(
            all_data_received=self.all_data_received,
            terminated=self.terminated,
        )

        self.protocol.makeConnection(self.transport)

    def assertStandardInputAndOutputClosed(self):
        """
        The transport's standard in, out, and error are closed.
        """
        self.assertTrue(self.transport_state.standard_in_closed)
        self.assertTrue(self.transport_state.standard_out_closed)
        self.assertTrue(self.transport_state.standard_error_closed)

    def test_receive_http_description(self):
        """
        Receiving a serialized :py:class:`_HTTPBinDescription` fires the
        ``all_data_received`` :py:class:`Deferred`.
        """
        self.assertNoResult(self.all_data_received)

        description = shared._HTTPBinDescription("host", 1234, "cert")

        self.protocol.lineReceived(
            json.dumps(attr.asdict(description)).encode('ascii')
        )

        self.assertStandardInputAndOutputClosed()

        self.assertEqual(self.successResultOf(self.all_data_received),
                         description)

    def test_receive_unexpected_line(self):
        """
        Receiving a line after the description synchronously raises in
        :py:class:`RuntimeError`
        """
        self.test_receive_http_description()
        with self.assertRaises(RuntimeError):
            self.protocol.lineReceived(b"unexpected")

    def test_connection_lost_before_receiving_data(self):
        """
        If the process terminates before its data is received, both
        ``all_data_received`` and ``terminated`` errback.
        """
        self.assertNoResult(self.all_data_received)

        self.protocol.connectionLost(Failure(ConnectionDone("done")))

        self.assertIsInstance(
            self.failureResultOf(self.all_data_received).value,
            ConnectionDone,
        )

        self.assertIsInstance(
            self.failureResultOf(self.terminated).value,
            ConnectionDone,
        )

    def test_connection_lost(self):
        """
        ``terminated`` fires when the connection is lost.
        """
        self.test_receive_http_description()

        self.protocol.connectionLost(Failure(ConnectionDone("done")))

        self.assertIsInstance(
            self.failureResultOf(self.terminated).value,
            ConnectionDone,
        )


@attr.s
class SpawnedProcess(object):
    """
    A call to :py:class:`MemoryProcessReactor.spawnProcess`.
    """
    process_protocol = attr.ib()
    executable = attr.ib()
    args = attr.ib()
    env = attr.ib()
    path = attr.ib()
    uid = attr.ib()
    gid = attr.ib()
    use_pty = attr.ib()
    child_fds = attr.ib()
    returned_process_transport = attr.ib()
    returned_process_transport_state = attr.ib()

    def send_stdout(self, data):
        """
        Send data from the process' standard out.

        :param data: The standard out data.
        """
        self.process_protocol.childDataReceived(1, data)

    def end_process(self, reason):
        """
        End the process.

        :param reason: The reason.
        :type reason: :py:class:`Failure`
        """
        self.process_protocol.processEnded(reason)


@implementer(IReactorCore, IReactorProcess)
class MemoryProcessReactor(MemoryReactor):
    """
    A fake :py:class:`IReactorProcess` and :py:class:`IReactorCore`
    provider to be used in tests.
    """
    def __init__(self):
        MemoryReactor.__init__(self)
        self.spawnedProcesses = []

    def spawnProcess(self, processProtocol, executable, args=(), env={},
                     path=None, uid=None, gid=None, usePTY=0, childFDs=None):
        """
        :ivar process_protocol: Stores the protocol passed to the reactor.
        :return: An L{IProcessTransport} provider.
        """
        transport_state = FakeProcessTransportState()
        transport = FakeProcessTransport(transport_state)

        self.spawnedProcesses.append(SpawnedProcess(
            process_protocol=processProtocol,
            executable=executable,
            args=args,
            env=env,
            path=path,
            uid=uid,
            gid=gid,
            use_pty=usePTY,
            child_fds=childFDs,
            returned_process_transport=transport,
            returned_process_transport_state=transport_state,
        ))

        processProtocol.makeConnection(transport)

        return transport


class MemoryProcessReactorTests(SynchronousTestCase):
    """
    Tests for :py:class:`MemoryProcessReactor`
    """

    def test_provides_interfaces(self):
        """
        :py:class:`MemoryProcessReactor` instances provide
        :py:class:`IReactorCore` and :py:class:`IReactorProcess`.
        """
        reactor = MemoryProcessReactor()
        verify.verifyObject(IReactorCore, reactor)
        verify.verifyObject(IReactorProcess, reactor)


class HTTPBinProcessTests(SynchronousTestCase):
    """
    Tests for :py:class:`_HTTPBinProcesss`.
    """

    def setUp(self):
        self.reactor = MemoryProcessReactor()
        self.opened_file_descriptors = []

    def fd_recording_open(self, *args, **kwargs):
        """
        Record the file descriptors of files opened by
        :py:func:`open`.

        :return: A file object.
        """
        fobj = open(*args, **kwargs)
        self.opened_file_descriptors.append(fobj.fileno())
        return fobj

    def spawned_process(self):
        """
        Assert that ``self.reactor`` has spawned only one process and
        return the :py:class:`SpawnedProcess` representing it.

        :return: The :py:class:`SpawnedProcess`.
        """
        self.assertEqual(len(self.reactor.spawnedProcesses), 1)
        return self.reactor.spawnedProcesses[0]

    def assertSpawnAndDescription(self, process, args, description):
        """
        Assert that spawning the given process invokes the command
        with the given args, that standard error is redirected, that
        it is killed at reactor shutdown, and that it returns a
        description that matches the provided one.

        :param process: :py:class:`_HTTPBinProcesss` instance.
        :param args: The arguments with which to execute the child
            process.
        :type args: :py:class:`tuple` of :py:class:`str`

        :param description: The expected
            :py:class:`_HTTPBinDescription`.

        :return: The returned :py:class:`_HTTPBinDescription`
        """
        process._open = self.fd_recording_open

        description_deferred = process.server_description(self.reactor)

        spawned_process = self.spawned_process()

        self.assertEqual(spawned_process.args, args)

        self.assertEqual(len(self.opened_file_descriptors), 1)
        [error_log_fd] = self.opened_file_descriptors

        self.assertEqual(spawned_process.child_fds.get(2), error_log_fd)

        self.assertNoResult(description_deferred)

        spawned_process.send_stdout(description.to_json_bytes() + b'\n')

        before_shutdown = self.reactor.triggers["before"]["shutdown"]
        self.assertEqual(len(before_shutdown), 1)
        [(before_shutdown_function, _, _)] = before_shutdown

        self.assertEqual(before_shutdown_function, process.kill)

        self.assertEqual(self.successResultOf(description_deferred),
                         description)

    def test_server_description_spawns_process(self):
        """
        :py:class:`_HTTPBinProcess.server_description` spawns an
        ``httpbin`` child process that it monitors with
        :py:class:`_HTTPBinServerProcessProtocol`, and redirects its
        standard error to a log file.
        """
        httpbin_process = parent._HTTPBinProcess(https=False)
        description = shared._HTTPBinDescription(host="host", port=1234)

        self.assertSpawnAndDescription(
            httpbin_process,
            [
                sys.executable,
                '-m',
                'treq.test.local_httpbin.child'
            ],
            description)

    def test_server_description_spawns_process_https(self):
        """
        :py:class:`_HTTPBinProcess.server_description` spawns an
        ``httpbin`` child process that listens over HTTPS, that it
        monitors with :py:class:`_HTTPBinServerProcessProtocol`, and
        redirects the process' standard error to a log file.
        """
        httpbin_process = parent._HTTPBinProcess(https=True)
        description = shared._HTTPBinDescription(host="host",
                                                 port=1234,
                                                 cacert="cert")

        self.assertSpawnAndDescription(
            httpbin_process,
            [
                sys.executable,
                '-m',
                'treq.test.local_httpbin.child',
                '--https',
            ],
            description)

    def test_server_description_caches_description(self):
        """
        :py:class:`_HTTPBinProcess.server_description` spawns an
        ``httpbin`` child process only once, after which it returns a
        cached :py:class:`_HTTPBinDescription`.
        """
        httpbin_process = parent._HTTPBinProcess(https=False)

        description_deferred = httpbin_process.server_description(self.reactor)

        self.spawned_process().send_stdout(
            shared._HTTPBinDescription(host="host", port=1234).to_json_bytes()
            + b'\n'
        )

        description = self.successResultOf(description_deferred)

        cached_description_deferred = httpbin_process.server_description(
            self.reactor,
        )

        cached_description = self.successResultOf(cached_description_deferred)

        self.assertIs(description, cached_description)

    def test_kill_before_spawn(self):
        """
        Killing a process before it has been spawned has no effect.
        """
        parent._HTTPBinProcess(https=False).kill()

    def test_kill(self):
        """
        Kill terminates the process as quickly as the platform allows,
        and the termination failure is suppressed.
        """
        httpbin_process = parent._HTTPBinProcess(https=False)

        httpbin_process.server_description(self.reactor)

        spawned_process = self.spawned_process()

        spawned_process.send_stdout(
            shared._HTTPBinDescription(host="host", port=1234).to_json_bytes()
            + b'\n'
        )

        termination_deferred = httpbin_process.kill()

        self.assertEqual(
            spawned_process.returned_process_transport_state.signals,
            ['KILL'],
        )

        spawned_process.end_process(
            Failure(ProcessTerminated(1, signal=signal.SIGKILL)),
        )

        self.successResultOf(termination_deferred)

    def test_kill_unexpected_exit(self):
        """
        The :py:class:`Deferred` returned by
        :py:meth:`_HTTPBinProcess.kill` errbacks with the failure when
        it is not :py:class:`ProcessTerminated`, or its signal does
        not match the expected signal.
        """
        for error in [ProcessTerminated(1, signal=signal.SIGIO),
                      ConnectionDone("Bye")]:
            httpbin_process = parent._HTTPBinProcess(https=False)

            httpbin_process.server_description(self.reactor)

            spawned_process = self.reactor.spawnedProcesses[-1]

            spawned_process.send_stdout(
                shared._HTTPBinDescription(host="host",
                                           port=1234).to_json_bytes()
                + b'\n'
            )

            termination_deferred = httpbin_process.kill()

            spawned_process.end_process(Failure(error))

            self.assertIs(self.failureResultOf(termination_deferred).value,
                          error)
